@@ -1,11 +1,11 @@
-/*  $Id: url.pl,v 1.53 2007/04/12 14:53:57 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -38,90 +38,43 @@
 	    http_location/2,		% ?Parts, ?Location
 	    www_form_encode/2,		% Value <-> Encoded
 	    parse_url_search/2,		% Form-data <-> Form fields
+	    
+	    url_iri/2,			% ?URL, ?IRI
 
-	    file_name_to_url/2		% ?FileName, ?URL
+	    file_name_to_url/2,		% ?FileName, ?URL
+
+	    set_url_encoding/2		% ?Old, +New
 	  ]).
 :- use_module(library(lists)).
+:- use_module(library(error)).
 :- use_module(library(utf8)).
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Utility library to break down URL   specifications. Originally this file
-was based on RFC-1738. In SWI-Prolog 5.6.13 it was upgraded to RFC 3986,
-notably covering UTF-8 encoding of Unicode characters in form-encoding.
+/** <module> Analysing and constructing URL
 
-Public interface:
+This library deals with the analysis and construction of a URL,
+Universal Resource Locator. URL is the basis for communicating locations
+of resources (data) on the web. A URL consists of a protocol identifier
+(e.g. HTTP, FTP, and a protocol-specific syntax further defining the
+location. URLs are standardized in RFC-1738.
 
-parse_url(+URL, -Parts)
-    Parse a URL into its parts.  Parts is a list of Name(Value).  Attributes
-    depend on the protocol.  Currently the implementation is not defined
-    for all protocols.  Defined:
-
-        protocol	All protocols
-	user		Network based protocols
-	host		Network based protocols
-	port		Network based protocols
-	path		file,http,ftp
-	search		http (?name=value&...)
-        fragment	http (#fragment)
-
-    For example:
-
-    ?- parse_url('http://www.swi-prolog.org/message.cgi?msg=Hello+World%21', P).
-
-    P = [ protocol(http), 
-	  search([msg='Hello World!']), 
-	  path('/message.cgi'), 
-	  host('www.swi-prolog.org')
-        ] 
-
-parse_url(-URL, +Parts)
-    Create a URL from its parts (see above). 	
-
-parse_url(+URL, +BaseURL, -Parts)
-    As above, but `URL' may be relative to some base url.  If URL is
-    absolute, BaseURL is ignored.
-
-parse_url(-URL, +BaseURL, +Parts)
-    Construct absolute URL from its parts and a base url, with default
-    parts taken from BaseURL. 
-
-global_url(+URL, +BaseURL, -AbsoluteUrl)
-    Transform a possible local URL into a global one.
-
-http_location(?Parts, ?Location)
-    If `Parts' is the part-list of an http URL, construct a string for
-    the location.  So, to access a page with an HTTP url, do:
-
-    	parse_url(URL, Parts),
-	memberchk(host(Host), Parts),
-	(   memberchk(port(Port), Parts)
-	->  true
-	;   Port = 80
-	),
-	<connect to host and port>
-	http_location(Parts, Location),
-	format(Stream, 'GET ~w HTTP/1.0\r\nHost: ~w\r\n\r\n',
-	       [ Location, Host ]),
-	...
-
-ISSUES:
-
-	* Lacks support for IP6 in parsing URIs
-	* Constructing could avoid percent encoding for some more
-	  characters
-	* Using + for spaces as used in forms may be active in too
-	  many places.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+The implementation in this library covers only a small portion of the
+defined protocols.  Though the initial implementation followed RFC-1738
+strictly, the current is more relaxed to deal with frequent violations
+of the standard encountered in practical use.
 
+@author	Jan Wielemaker
+@author Lukas Faulstich
+*/
 
 		 /*******************************
 		 *	      GLOBALISE		*
 		 *******************************/
 
-%%	global_url(+URL, +Base, -Global)
+%%	global_url(+URL, +Base, -Global) is det.
 %	
-%	Translate a relative URL into an absolute one.  The first three
-%	cases deal with commonly seen and quickly to resolve cases.
+%	Translate a possibly relative URL  into   an  absolute  one.
+%	
+%	@error syntax_error(illegal_url) if URL is not legal.
 
 global_url(URL, BaseURL, Global) :-
 	(   is_absolute_url(URL),
@@ -137,15 +90,16 @@ global_url(URL, BaseURL, Global) :-
 		atom_concat(BaseURL, NoHash, Global)
 	    ;	atom_concat(BaseURL, URL, Global)
 	    )
-	;   parse_url(URL, BaseURL, Attributes),
-	    phrase(curl(Attributes), Chars),
+	;   parse_url(URL, BaseURL, Attributes)
+	->  phrase(curl(Attributes), Chars),
 	    atom_codes(Global, Chars)
+	;   throw(error(syntax_error(illegal_url), URL))
 	).
 
 %%	is_absolute_url(+URL)
 %	
-%	Test whether a URL is absolute or relative.  We assume it is
-%	absolute if it starts with a protocol.
+%	True if URL is an absolute URL. That  is, a URL that starts with
+%	a protocol identifier.
 
 is_absolute_url(URL) :-
 	sub_atom(URL, 0, _, _, 'http://'), !.
@@ -166,8 +120,16 @@ is_absolute_url(URL) :-
 
 %%	http_location(?Parts, ?Location)
 %
-%	Translate the relevant parts of an URL into an HTTP location
-%	on a server.
+%	Construct or analyze an  HTTP  location.   This  is  similar  to
+%	parse_url/2, but only deals with the   location  part of an HTTP
+%	URL. That is, the path, search   and fragment specifiers. In the
+%	HTTP protocol, the first line of a message is
+%	
+%	    ==
+%	    <Action> <Location> HTTP/<version>
+%	    ==
+%	    
+%	@param Location	Atom or list of character codes.    
 
 http_location(Parts, Location) :-	% Parts --> Location
 	nonvar(Parts), !,
@@ -311,18 +273,61 @@ cfragment(_) -->
 %%	parse_url(+URL, -Attributes) is det.
 %%	parse_url(+URL, +BaseURL, -Attributes) is det.
 %
-%	Parse a URL to a sequence  of   attributes.  In the version with
-%	three  arguments,  URL  may  be  an  URL  relative  to  BaseURL.
-%	Attributes:
-%
-%		| protocol(Protocol)	| Protocol identifier |
-%		| host(Host)		| Name of host |
-%		| port(Port)		| Number of port to contact |
-%		| path(Path)		| The path |
-%		| search(Search)	| Search specification |
-%		
-%	The parse_url/3	predicate deals with relative URLs.  BaseURL can
-%	be specified as an atom or parsed URL.
+%	Construct or analyse a URL. URL is an   atom  holding a URL or a
+%	variable. Parts is a list of   components.  Each component is of
+%	the format Name(Value). Defined components are:
+%	
+%	    * protocol(Protocol)
+%	    The used protocol. This is, after  the optional =|url:|=, an
+%	    identifier separated from the remainder of  the URL using :.
+%	    parse_url/2 assumes the =http= protocol   if  no protocol is
+%	    specified and the URL can be parsed  as a valid HTTP url. In
+%	    addition to the RFC-1738  specified   protocols,  the =file=
+%	    protocol is supported as well.
+%	    
+%	    * host(Host)
+%	    Host-name or IP-address on which   the  resource is located.
+%	    Supported by all network-based protocols.
+%	    
+%	    * port(Port)
+%	    Integer port-number to access on   the \arg{Host}. This only
+%	    appears if the port is  explicitly   specified  in  the URL.
+%	    Implicit default ports (e.g.  80   for  HTTP)  do \emph{not}
+%	    appear in the part-list.
+%	    
+%	    * path(Path)
+%	    (File-) path addressed by the URL. This is supported for the
+%	    =ftp=, =http= and =file= protocols. If  no path appears, the
+%	    library generates the path =|/|=.
+%	    
+%	    * search(ListOfNameValue)
+%	    Search-specification of HTTP URL. This is the part after the
+%	    =|?|=, normally used to transfer data   from HTML forms that
+%	    use the =GET=  protocol.  In  the   URL  it  consists  of  a
+%	    www-form-encoded list of Name=Value pairs. This is mapped to
+%	    a list of Prolog Name=Value  terms   with  decoded names and
+%	    values.
+%	    
+%	    * fragment(Fragment)
+%	    Fragment specification of HTTP URL. This   is the part after
+%	    the =|#|= character.
+%	    
+%	The example below illustrates the all this for an HTTP URL.
+%	
+%	    ==
+%	    ?- parse_url('http://swi.psy.uva.nl/message.cgi?msg=Hello+World%21#x', P).
+%	    
+%	    P = [ protocol(http),
+%		  host('swi.psy.uva.nl'),
+%		  fragment(x),
+%		  search([ msg = 'Hello World!'
+%			 ]),
+%		  path('/message.cgi')
+%	        ]
+%	    ==
+%	    
+%	By instantiating the parts-list this predicate   can  be used to
+%	create a URL.
 
 parse_url(URL, Attributes) :-
 	nonvar(URL), !,
@@ -393,9 +398,16 @@ make_path(Dir, Local, Path) :-
 	concat_atom([Dir, /, Local], Path).
 
 
+%%	absolute_url//
+%
+%	True if the input  describes  an   absolute  URL.  This means it
+%	starts with a URL schema. We demand a   schema  of length > 1 to
+%	avoid confusion with Windows drive letters.
+
 absolute_url -->
-	lwalpha(_),
-	schema_chars(_),
+	lwalpha(_First),
+	schema_chars(Rest),
+	{ Rest \== [] },
 	":", !.
 
 
@@ -424,7 +436,7 @@ digit(C, [C|T], T) :- code_type(C, digit).
 url([protocol(Schema)|Parts]) -->
 	schema(Schema),
 	":", !,
-	hier_part(Parts, P2),
+	hier_part(Schema, Parts, P2),
 	query(P2, P3),
 	fragment(P3, []).
 url([protocol(http)|Parts]) -->		% implicit HTTP
@@ -489,14 +501,19 @@ schema_extra(0'.).	% 0'
 
 
 %%	hier_part(+Schema, -Parts, ?Tail)//
-%	
-%	Extract the hierarchy part.
 
-hier_part(Parts, Tail) -->
+hier_part(file, [path(Path)|Tail], Tail) --> !,
+	"//",
+	(   win_drive_path(Path)
+	;   path_absolute(Path)
+	;   path_rootless(Path)
+	;   path_empty(Path)
+	), !.
+hier_part(_, Parts, Tail) -->
 	"//", !,
 	authority(Parts, [path(Path)|Tail]),
 	path_abempty(Path).
-hier_part([path(Path)|T], T) -->
+hier_part(_, [path(Path)|T], T) -->
 	(   path_absolute(Path)
 	;   path_rootless(Path)
 	;   path_empty(Path)
@@ -574,6 +591,18 @@ path_abempty(Path) -->
 	}.
 
 
+win_drive_path(Path) -->
+	drive_letter(C0),
+	":",
+	(   "/"
+	->  {Codes = [C0, 0':, 0'/|Chars]}
+	;   {Codes = [C0, 0':|Chars]}
+	),
+	segment_nz_chars(Chars, T0),
+	segments_chars(T0, []),
+	{ atom_codes(Path, Codes) }.
+
+
 path_absolute(Path) -->
 	"/",
 	segment_nz_chars(Chars, T0),
@@ -625,7 +654,7 @@ segment_nz_nc_char(C) --> pchar(C).
 %	Extract &Name=Value, ...
 
 query([search(Params)|T], T) -->
-	"?",
+	"?", !,
 	search(Params).
 query(T,T) -->
 	[].
@@ -687,6 +716,8 @@ fragment_chars([]) -->
 fragment_char(C)   --> pchar(C), !.
 fragment_char(0'/) --> "/", !.
 fragment_char(0'?) --> "?", !.
+fragment_char(0'[) --> "[", !.		% Not according RDF3986!
+fragment_char(0']) --> "]", !.
 
 
 		 /*******************************
@@ -699,7 +730,7 @@ fragment_char(0'?) --> "?", !.
 %	
 %	Performs UTF-8 decoding of percent encoded strings.
 
-pchar(0' ) --> "+", !.			% ?
+pchar(0' ) --> "+", !.			%' ?
 pchar(C) -->
 	[C],
 	{   unreserved(C)
@@ -721,6 +752,13 @@ lwalpha(H) -->
 	  code_type(H, to_lower(C))
 	}.
 
+drive_letter(C) -->
+	[C],
+	{ C < 128,
+	  code_type(C, alpha)
+	}.
+
+
 		 /*******************************
 		 *	RESERVED CHARACTERS	*
 		 *******************************/
@@ -747,13 +785,18 @@ sub_delim(0'=).
 %	Characters that can be represented without procent escaping
 %	RFC 3986, section 2.3
 
-unreserved(C) :-
-	code_type(C, alnum),
-	C < 128.
-unreserved(0'-).
-unreserved(0'.).
-unreserved(0'_).
-unreserved(0'~).			% 0'
+term_expansion(unreserved(map), Clauses) :-
+	findall(unreserved(C), unreserved_(C), Clauses).
+
+unreserved_(C) :-
+	between(1, 128, C),
+	code_type(C, alnum).
+unreserved_(0'-).
+unreserved_(0'.).
+unreserved_(0'_).
+unreserved_(0'~).			% 0'
+
+unreserved(map).			% Expanded
 
 
 		 /*******************************
@@ -769,8 +812,8 @@ encoding used with the HTTP GET.
 %%	www_form_encode(-Value, +XWWWFormEncoded) is det.
 %
 %	En/Decode        between        native          value        and
-%	=application/x-www-form-encoded=. Maps space to  +, keeps alnum,
-%	maps anything else to  =%XX=  and   newlines  to  =%OD%OA.= When
+%	application/x-www-form-encoded. Maps space to   +,  keeps alnum,
+%	maps anything else to =|%XX|= and   newlines to =|%OD%OA|=. When
 %	decoding, newlines appear as a single newline (10) character.
 
 www_form_encode(Value, Encoded) :-
@@ -797,18 +840,23 @@ www_encode([H|T], Extra) -->
 www_encode([], _) -->
 	"".
 
-percent_encode(C, Extra) -->
-	{ unreserved(C)
-	; memberchk(C, Extra)
-	}, !,
+percent_encode(C, _Extra) -->
+	{ unreserved(C) }, !,
 	[C].
-percent_encode(0' , _) --> !, "+".
+percent_encode(C, Extra) -->
+	{ memberchk(C, Extra) }, !,
+	[C].
+%percent_encode(0' , _) --> !, "+".	% Deprecated: use %20
 percent_encode(C, _) -->
 	{ C =< 128 }, !,
 	percent_byte(C).
 percent_encode(C, _) -->		% Unicode characters
-	{ phrase(utf8_codes([C]), Bytes) },
+	{ current_prolog_flag(url_encoding, utf8), !,
+	  phrase(utf8_codes([C]), Bytes)
+	},
 	percent_bytes(Bytes).
+percent_encode(C, _) -->
+	percent_byte(C).
 
 percent_bytes([]) -->
 	"".
@@ -866,16 +914,71 @@ utf8_cont([]) -->
 	[].
 
 
+%%	set_url_encoding(?Old, +New) is semidet.
+%
+%	Query and set the encoding for URLs.  The default is =utf8=.
+%	The only other defined value is =iso_latin_1=.
+%	
+%	@tbd	Having a global flag is highly inconvenient, but a 
+%		work-around for old sites using ISO Latin 1 encoding.
+
+:- set_prolog_flag(url_encoding, utf8).
+
+set_url_encoding(Old, New) :-
+	current_prolog_flag(url_encoding, Old),
+	(   Old == New
+	->  true
+	;   must_be(oneof([utf8, iso_latin_1]), New),
+	    set_prolog_flag(url_encoding, New)
+	).
+
+
+		 /*******************************
+		 *	 IRI PROCESSING		*
+		 *******************************/
+
+%%	url_iri(+Encoded, -Decoded) is det.
+%%	url_iri(-Encoded, +Decoded) is det.
+%	
+%	Convert between a URL, encoding in US-ASCII   and an IRI. An IRI
+%	is a fully expanded Unicode string.   Unicode  strings are first
+%	encoded into UTF-8, after which %-encoding takes place.
+
+url_iri(Encoded, Decoded) :-
+	nonvar(Encoded), !,
+	(   sub_atom(Encoded, _, _, _, '%')
+	->  atom_codes(Encoded, Codes),
+	    unescape_precent(Codes, UTF8),
+	    phrase(utf8_codes(Unicodes), UTF8),
+	    atom_codes(Decoded, Unicodes)
+	;   Decoded = Encoded
+	).
+url_iri(URL, IRI) :-
+	atom_codes(IRI, IRICodes),
+	phrase(percent_encode(IRICodes, "/:?#&="), UrlCodes),
+	atom_codes(URL, UrlCodes).
+
+
+unescape_precent([], []).
+unescape_precent([0'%,C1,C2|T0], [H|T]) :- !,	%'
+	code_type(C1, xdigit(D1)),
+	code_type(C2, xdigit(D2)),
+	H is D1*16 + D2,
+	unescape_precent(T0, T).
+unescape_precent([H|T0], [H|T]) :-
+	unescape_precent(T0, T).
+
 
 		 /*******************************
 		 *	     FORM DATA		*
 		 *******************************/
 
-%%	parse_url_search(?Spec, ?Fields) is det.
+%%	parse_url_search(?Spec, ?Fields:list(Name=Value)) is det.
 %
-%	Parse  between  a  list   of    Name=Value   and  the  MIME-type
-%	=application/x-www-form-urlencoded=  as  used  to    post   HTTP
-%	requests
+%	Construct or analyze an HTTP   search  specification. This deals
+%	with       form       data       using       the       MIME-type
+%	=application/x-www-form-urlencoded=  as  used   in    HTTP   GET
+%	requests.
 
 parse_url_search(Spec, Fields) :-
 	atomic(Spec), !,
@@ -895,9 +998,10 @@ parse_url_search(Codes, Fields) :-
 %%	file_name_to_url(+File, -URL) is det.
 %%	file_name_to_url(-File, +URL) is semidet.
 %
-%	Translate between a filename and a URL.
+%	Translate between a filename and a file:// URL.
 %	
-%	@tbd	Deal with encoding?
+%	@tbd	Current implementation does not deal with paths that
+%		need special encoding.
 
 file_name_to_url(File, FileURL) :-
 	nonvar(File), !,
