@@ -185,7 +185,7 @@ http_status_reply(moved(To), Out, HrdExtra) :- !,
 		      p(['The document has moved ',
 			 a(href(To), ' Here')
 			]),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(moved(To, HTML), HrdExtra), Header),
@@ -198,7 +198,7 @@ http_status_reply(moved_temporary(To), Out, HrdExtra) :- !,
 		      p(['The document is currently ',
 			 a(href(To), ' Here')
 			]),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(moved_temporary(To, HTML), HrdExtra), Header),
@@ -211,7 +211,7 @@ http_status_reply(see_other(To),Out,HdrExtra) :- !,
                       p(['See other document ',
                          a(href(To), ' Here')
                         ]),
-                      address(httpd)
+                      \address
                     ]),
                HTML),
         phrase(reply_header(see_other(To, HTML), HdrExtra), Header),
@@ -224,7 +224,7 @@ http_status_reply(not_found(URL), Out, HrdExtra) :- !,
 		      p(['The requested URL ', tt(URL),
 			 ' was not found on this server'
 			]),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(status(not_found, HTML), HrdExtra), Header),
@@ -237,7 +237,7 @@ http_status_reply(forbidden(URL), Out, HrdExtra) :- !,
 		      p(['You don\'t have permission to access ', URL,
 			 ' on this server'
 			]),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(status(forbidden, HTML), HrdExtra), Header),
@@ -254,59 +254,44 @@ http_status_reply(authorise(Method, Realm), Out, HrdExtra) :- !,
 			 'browser doesn\'t understand how to supply ',
 			 'the credentials required.'
 			]),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(authorise(Method, Realm, HTML), HrdExtra), Header),
 	format(Out, '~s', [Header]),
 	print_html(Out, HTML).
 http_status_reply(not_modified, Out, HrdExtra) :- !,
-	phrase(page([ title('304 Not Modified')
-		    ],
-		    [ h1('Not Modified'),
-		      p(['The resource has not changed']),
-		      address(httpd)
-		    ]),
-	       HTML),
-	phrase(reply_header(status(not_modified, HTML), HrdExtra), Header),
-	format(Out, '~s', [Header]),
-	print_html(Out, HTML).
+	phrase(reply_header(status(not_modified), HrdExtra), Header),
+	format(Out, '~s', [Header]).
 http_status_reply(server_error(ErrorTerm), Out, HrdExtra) :-
 	'$messages':translate_message(ErrorTerm, Lines, []),
 	phrase(page([ title('500 Internal server error')
 		    ],
 		    [ h1('Internal server error'),
 		      p(\html_message_lines(Lines)),
-		      address(httpd)
+		      \address
 		    ]),
 	       HTML),
 	phrase(reply_header(status(server_error, HTML), HrdExtra), Header),
 	format(Out, '~s', [Header]),
 	print_html(Out, HTML).
-http_status_reply(resource_error(ErrorTerm), Out, HrdExtra) :- !,
+http_status_reply(unavailable(WhyHTML), Out, HdrExtra) :- !,
+	phrase(page([ title('503 Service Unavailable')
+		    ],
+		    [ h1('Service Unavailable'),
+		      WhyHTML,
+		      \address
+		    ]),
+	       HTML),
+	phrase(reply_header(status(service_unavailable, HTML), HdrExtra), Header),
+	format(Out, '~s', [Header]),
+	print_html(Out, HTML).
+http_status_reply(resource_error(ErrorTerm), Out, HdrExtra) :- !,
 	'$messages':translate_message(ErrorTerm, Lines, []),
-	phrase(page([ title('503 Service Unavailable')
-		    ],
-		    [ h1('Service Unavailable'),
-		      p(['The server is temporarily out of resources, please try again later']),
-		      p(\html_message_lines(Lines)),
-		      address(httpd)
-		    ]),
-	       HTML),
-	phrase(reply_header(status(service_unavailable, HTML), HrdExtra), Header),
-	format(Out, '~s', [Header]),
-	print_html(Out, HTML).
-http_status_reply(busy, Out, HrdExtra) :- !,
-	phrase(page([ title('503 Service Unavailable')
-		    ],
-		    [ h1('Service Unavailable'),
-		      p(['The server is temporarily out of resources, please try again later']),
-		      address(httpd)
-		    ]),
-	       HTML),
-	phrase(reply_header(status(service_unavailable, HTML), HrdExtra), Header),
-	format(Out, '~s', [Header]),
-	print_html(Out, HTML).
+	http_status_reply(unavailable(p(\html_message_lines(Lines))), Out, HdrExtra).
+http_status_reply(busy, Out, HdrExtra) :- !,
+	HTML = p(['The server is temporarily out of resources, please try again later']),
+	http_status_reply(unavailable(HTML), Out, HdrExtra).
 
 
 html_message_lines([]) -->
@@ -727,6 +712,10 @@ reply_header(see_other(To,Tokens),HdrExtra) -->
 	content_length(html(Tokens)),
 	content_type(text/html, utf8),
 	"\r\n".
+reply_header(status(Status), HdrExtra) --> % Empty messages: 1xx, 204 and 304
+	vstatus(Status),
+	header_fields(HdrExtra),
+	"\r\n".
 reply_header(status(Status, Tokens), HdrExtra) -->
 	vstatus(Status),
 	date(now),
@@ -1223,6 +1212,27 @@ header([]) -->
 	blanks,
 	[].
 
+%%	address//
+%
+%	Emit the HTML for the server address on behalve of error and
+%	status messages (non-200 replies).  Default is
+%
+%	    ==
+%	    SWI-Prolog httpd at <hostname>
+%	    ==
+%
+%	The address can be modified by   providing  a definition for the
+%	multifile predicate http:http_address//0.
+
+:- multifile
+	http:http_address//0.
+
+address -->
+	http:http_address, !.
+address -->
+	{ gethostname(Host) },
+	html(address(['SWI-Prolog httpd at ', Host])).
+
 mkfield(host, Host:Port, [host(Host),port(Port)|Tail], Tail) :- !.
 mkfield(Name, Value, [Att|Tail], Tail) :-
 	Att =.. [Name, Value].
