@@ -1,11 +1,11 @@
-/*  $Id: http_open.pl,v 1.21 2007/06/20 08:58:33 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2007, University of Amsterdam
+    Copyright (C): 2008, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -31,7 +31,8 @@
 
 
 :- module(http_open,
-	  [ http_open/3			% +URL, -Stream, +Options
+	  [ http_open/3,		% +URL, -Stream, +Options
+	    http_set_authorization/2	% +URL, +Authorization
 	  ]).
 :- use_module(library(url)).
 :- use_module(library(readutil)).
@@ -39,8 +40,10 @@
 :- use_module(library(lists)).
 :- use_module(library(option)).
 :- use_module(library(error)).
+:- use_module(library(base64)).
+:- use_module(library(debug)).
 
-user_agent('SWI-Prolog (http://www.swi-prolog.org)').
+user_agent('SWI-Prolog <http://www.swi-prolog.org>').
 
 /** <module> Simple HTTP client
 
@@ -73,6 +76,8 @@ client support is provided by http_client.pl
 %		  Use an HTTP proxy server
 %		* user_agent(+Agent)
 %		  User agent for identifying
+%		* authorization(+Term)
+%		  Currently Term is basic(User, Password)
 %		* request_header(Name=Value)
 %		  Extra header field
 %		* method(+Method)
@@ -85,25 +90,28 @@ client support is provided by http_client.pl
 
 http_open(URL, Stream, Options) :-
 	atom(URL), !,
-	parse_url(URL, Parts),
-	http_open(Parts, Stream, Options).
+	parse_url_ex(URL, Parts),
+	add_authorization(URL, Options, Options1),
+	http_open(Parts, Stream, Options1).
 http_open(Parts, Stream, Options0) :-
+	memberchk(proxy(Host, ProxyPort), Options0), !,
+	parse_url_ex(Location, Parts),
 	Options = [visited(Parts)|Options0],
-	memberchk(proxy(Host, ProxyPort), Options), !,
-	parse_url(Location, Parts),
 	open_socket(Host:ProxyPort, In, Out, Options),
 	option(port(Port), Parts, 80),
 	host_and_port(Host, Port, HostPort),
-	send_rec_header(Out, In, Stream, HostPort, Location, Parts, Options),
+	add_authorization(Parts, Options, Options1),
+	send_rec_header(Out, In, Stream, HostPort, Location, Parts, Options1),
 	return_final_url(Options).
 http_open(Parts, Stream, Options0) :-
-	Options = [visited(Parts)|Options0],
 	memberchk(host(Host), Parts),
 	option(port(Port), Parts, 80),
 	http_location(Parts, Location),
+	Options = [visited(Parts)|Options0],
 	open_socket(Host:Port, In, Out, Options),
 	host_and_port(Host, Port, HostPort),
-	send_rec_header(Out, In, Stream, HostPort, Location, Parts, Options),
+	add_authorization(Parts, Options, Options1),
+	send_rec_header(Out, In, Stream, HostPort, Location, Parts, Options1),
 	return_final_url(Options).
 
 host_and_port(Host, 80, Host) :- !.
@@ -178,15 +186,25 @@ x_headers([H|T], Out) :- !,
 
 x_header(request_header(Name=Value), Out) :- !,
 	format(Out, '~w: ~w\r\n', [Name, Value]).
+x_header(authorization(Authorization), Out) :- !,
+	auth_header(Authorization, Out).
 x_header(_, _).
 
+auth_header(basic(User, Password), Out) :- !,
+	format(codes(Codes), '~w:~w', [User, Password]),
+	phrase(base64(Codes), Base64Codes),
+	format(Out, 'Authorization: basic ~s\r\n', [Base64Codes]).
+auth_header(Auth, _) :-
+	domain_error(authorization, Auth).
+
 user_agent(Agent, Options) :-
 	(   option(user_agent(Agent), Options)
 	->  true
 	;   user_agent(Agent)
 	).
 
-% %	do_open(+HTTPStatusCode, +HTTPStatusComment, +Header +Options, +Parts, +In, -FinalIn)
+%%	do_open(+HTTPStatusCode, +HTTPStatusComment, +Header,
+%%		+Options, +Parts, +In, -FinalIn) is det.
 %
 %	Handle the HTTP status. If 200, we   are ok. If a redirect, redo
 %	the open, returning a new stream. Else issue an error.
@@ -198,22 +216,38 @@ do_open(200, _, Lines, Options, Parts, In0, In) :- !,
 	return_fields(Options, Lines),
 	transfer_encoding_filter(Lines, In0, In),
 					% properly re-initialise the stream
-	parse_url(Id, Parts),
+	parse_url_ex(Id, Parts),
 	set_stream(In, file_name(Id)),
 	set_stream(In, record_position(true)).
 					% Handle redirections
 do_open(Code, _, Lines, Options, Parts, In, Stream) :-
 	redirect_code(Code),
 	location(Lines, Location), !,
-	parse_url(Location, Parts, Redirected),
+	parse_url_ex(Location, Parts, Redirected),
 	close(In),
 	http_open(Redirected, Stream, [visited(Redirected)|Options]).
 					% report anything else as error
-do_open(Code, Comment, _,  _, Parts, In, _) :-
-	close(In),
-	parse_url(Id, Parts),
-	throw(error(existence_error(url, Id),
-		    context(_, status(Code, Comment)))).
+do_open(Code, Comment, _,  _, Parts, _, _) :-
+	parse_url_ex(Id, Parts),
+	(   map_error_code(Code, Error)
+	->  Formal =.. [Error, url, Id]
+	;   Formal = existence_error(url, Id)
+	),
+	throw(error(Formal, context(_, status(Code, Comment)))).
+
+%%	map_error_code(+HTTPCode, -PrologError) is semidet.
+%
+%	Map HTTP error codes to Prolog errors.
+%	
+%	@tbd	Many more maps. Unfortunately many have no sensible Prolog
+%		counterpart.
+
+map_error_code(401, permission_error).
+map_error_code(403, permission_error).
+map_error_code(404, existence_error).
+map_error_code(405, permission_error).
+map_error_code(407, permission_error).
+map_error_code(410, existence_error).
 
 redirect_code(301).			% moved permanently
 redirect_code(302).			% moved temporary
@@ -257,7 +291,7 @@ return_final_url(Options) :-
 	memberchk(final_url(URL), Options),
 	var(URL), !,
 	memberchk(visited(Parts), Options),
-	parse_url(URL, Parts).
+	parse_url_ex(URL, Parts).
 return_final_url(_).
 
 	
@@ -282,6 +316,7 @@ transfer_encoding_filter(_, In, In).
 transfer_encoding(Lines, Encoding) :-
 	member(Line, Lines),
 	phrase(transfer_encoding(Encoding0), Line), !,
+	debug(http(transfer_encoding), 'Transfer-encoding: ~w', [Encoding0]),
 	Encoding = Encoding0.
 
 transfer_encoding(Encoding) -->
@@ -387,5 +422,84 @@ rest(A,L,[]) :-
 	atom_codes(A, L).
 
 
+		 /*******************************
+		 *   AUTHORIZATION MANAGEMENT	*
+		 *******************************/
+
+%%	http_set_authorization(+URL, +Authorization) is det.
+%
+%	Set user/password to supply with URLs   that have URL as prefix.
+%	If  Authorization  is  the   atom    =|-|=,   possibly   defined
+%	authorization is cleared.  For example:
+%	
+%	==
+%	?- http_set_authorization('http://www.example.com/private/',
+%				  basic('John', 'Secret'))
+%	==
+%	
+%	@tbd	Move to a separate module, so http_get/3, etc. can use this
+%		too.
 	
+:- dynamic
+	stored_authorization/2,
+	cached_authorization/2.
+
+http_set_authorization(URL, Authorization) :-
+	must_be(atom, URL),
+	retractall(stored_authorization(URL, _)),
+	(   Authorization = (-)
+	->  true
+	;   check_authorization(Authorization),
+	    assert(stored_authorization(URL, Authorization))
+	),
+	retractall(cached_authorization(_,_)).
+
+check_authorization(Var) :-
+	var(Var), !,
+	instantiation_error(Var).
+check_authorization(basic(User, Password)) :-
+	must_be(atom, User),
+	must_be(atom, Password).
+
+%%	authorization(+URL, -Authorization) is semdet.
+%
+%	True if Authorization must be supplied for URL.
+%	
+%	@tbd	Cleanup cache if it gets too big.
+
+authorization(_, _) :-
+	\+ stored_authorization(_, _), !,
+	fail.
+authorization(URL, Authorization) :-
+	cached_authorization(URL, Authorization), !,
+	Authorization \== (-).
+authorization(URL, Authorization) :-
+	(   stored_authorization(Prefix, Authorization),
+	    sub_atom(URL, 0, _, _, Prefix)
+	->  assert(cached_authorization(URL, Authorization))
+	;   assert(cached_authorization(URL, -)),
+	    fail
+	).
 	
+add_authorization(_, Options, Options) :-
+	option(authorization(_), Options), !.
+add_authorization(For, Options0, Options) :-
+	stored_authorization(_, _) ->	% quick test to avoid work
+	(   atom(For)
+	->  URL = For
+	;   parse_url_ex(URL, For)
+	),
+	authorization(URL, Auth), !,
+	Options = [authorization(Auth)|Options0].
+add_authorization(_, Options, Options).
+	
+
+parse_url_ex(URL, Parts) :-
+	parse_url(URL, Parts), !.
+parse_url_ex(URL, _) :-
+	domain_error(url, URL).		% Syntax error?
+
+parse_url_ex(URL, RelativeTo, Parts) :-
+	parse_url(URL, RelativeTo, Parts), !.
+parse_url_ex(URL, _, _) :-
+	domain_error(url, URL).		% Syntax error?
