@@ -1,4 +1,4 @@
-/*  $Id: doc_wiki.pl,v 1.8 2007/03/22 11:37:00 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -42,6 +42,7 @@
 :- use_module(library(lists)).
 :- use_module(library(debug)).
 :- use_module(library(memfile)).
+:- use_module(library(pairs)).
 
 
 		 /*******************************
@@ -96,6 +97,8 @@ take_par(Lines, List, Rest) :-
 	List0 =.. [Type, LI],
 	(   ul_to_dl(List0, List)
 	->  true
+	;   List0 = dl(Items)
+	->  List = dl(class=wiki, Items)
 	;   List = List0
 	).
 take_par([N-['|'|RL1]|LT], table(class=wiki, [tr(R0)|RL]), Rest) :-
@@ -133,6 +136,8 @@ list_item(Lines, _, Indent, [SubList|LIT], LIT, Rest) :-	% sub-list
 	take_par(Lines, SubList, Rest).
 
 %%	rest_list_item(+Lines, +Type, +Indent, -RestItem, -RestLines) is det
+%
+%	Extract the remainder (after the first line) of a list item.
 
 rest_list_item([], _, _, [], []).
 rest_list_item([_-[]|L], _, _, [], L) :- !.	% empty line
@@ -145,6 +150,7 @@ rest_list_item([_-L1|L0], Type, N, ['\n'|LI], L) :-
 	append(L1, LIT, LI),
 	rest_list_item(L0, Type, N, LIT, L).
 
+
 %%	rest_list(+Lines, +Type, +Indent,
 %%		  -Items, -ItemTail, -RestLines) is det.
 
@@ -153,7 +159,7 @@ rest_list(Lines, Type, N, Items, IT, Rest) :-
 	rest_list(Rest0, Type, N, IT0, IT, Rest).
 rest_list(Rest, _, _, IT, IT, Rest).
 
-%%	list_item_Line(?Type, +Line, -Rest) is det.
+%%	list_item_prefix(?Type, +Line, -Rest) is det.
 
 list_item_prefix(ul, [*, ' '|T], T) :- !.
 list_item_prefix(dl, [$, ' '|T], T) :-
@@ -165,11 +171,16 @@ list_item_prefix(ol, [N, '.', ' '|T], T) :-
 
 %%	ul_to_dl(+UL, -DL) is semidet.
 %
-%	Translate an UL list into a DL list if all entries are of the
-%	form "* <term> nl, <description>"
+%	Translate an UL list into a DL list   if  all entries are of the
+%	form "* <term> nl, <description>" or all   items  are of the for
+%	[[PredicateIndicator]].
 
-ul_to_dl(ul(Items), dl(class(termlist), DLItems)) :-
-	term_items(Items, DLItems, []).
+ul_to_dl(ul(Items), Description) :-
+	term_items(Items, DLItems, []),
+	(   terms_to_predicate_includes(DLItems, Preds)
+	->  Description = dl(class(predicates), Preds)
+	;   Description = dl(class(termlist), DLItems)
+	).
 
 term_items([], T, T).
 term_items([LI|LIs], DLItems, Tail) :-
@@ -187,7 +198,11 @@ term_item(li(Tokens),
 	    dd(Descr)
 	  | Tail
 	  ], Tail) :-
-	(   append(TermTokens, ['\n'|Descr], Tokens)
+	(   (   append(TermTokens, ['\n'|Descr], Tokens)
+	    ->	true
+	    ;	TermTokens = Tokens,
+		Descr = []
+	    )
 	->  new_memory_file(MemFile),
 	    open_memory_file(MemFile, write, Out),
 	    forall(member(T, TermTokens),
@@ -207,6 +222,21 @@ read_dt_term(In, Term, Bindings) :-
 		    module(pldoc_modes)
 		  ]).
 
+terms_to_predicate_includes([], []).
+terms_to_predicate_includes([dt(class=term, \term([[PI]], [])), dd([])|T0],
+			    [\include(PI, predicate)|T]) :-
+	is_pi(PI),
+	terms_to_predicate_includes(T0, T).
+
+is_pi(Name/Arity) :-
+	atom(Name),
+	integer(Arity),
+	between(0, 20, Arity).
+is_pi(Name//Arity) :-
+	atom(Name),
+	integer(Arity),
+	between(0, 20, Arity).
+
 
 %%	row(-Cells)// is det.
 
@@ -238,19 +268,32 @@ rest_par([_-L1|LT], ['\n'|Par], Rest) :-
 
 %%	section_line(+Tokens, -Section) is det.
 %
-%	Extract a section using the Wiki conventions
+%	Extract a section using the Twiki   conventions. The section may
+%	be preceeded by [Word], in which case we generate an anchor name
+%	Word for the section.
 
 section_line([-,-,-|Rest], Section) :-
 	plusses(Rest, Section).
 
-plusses([+, ' '|Rest], h1(class(wiki), Content)) :-
-	strip_ws_tokens(Rest, Content).
-plusses([+, +, ' '|Rest], h2(class(wiki), Content)) :-
-	strip_ws_tokens(Rest, Content).
-plusses([+, +, +, ' '|Rest], h3(class(wiki), Content)) :-
-	strip_ws_tokens(Rest, Content).
-plusses([+, +, +, +, ' '|Rest], h4(class(wiki), Content)) :-
-	strip_ws_tokens(Rest, Content).
+plusses([+, ' '|Rest], h1(Attrs, Content)) :-
+	hdr_attributes(Rest, Attrs, Content).
+plusses([+, +, ' '|Rest], h2(Attrs, Content)) :-
+	hdr_attributes(Rest, Attrs, Content).
+plusses([+, +, +, ' '|Rest], h3(Attrs, Content)) :-
+	hdr_attributes(Rest, Attrs, Content).
+plusses([+, +, +, +, ' '|Rest], h4(Attrs, Content)) :-
+	hdr_attributes(Rest, Attrs, Content).
+
+hdr_attributes(List, Attrs, Content) :-
+	strip_leading_ws(List, List2),
+	(   List2 = ['[',Word,']'|List3],
+	    atomic(Word)
+	->  strip_ws_tokens(List3, Content),
+	    string_to_atom(Word, Name),
+	    Attrs = [class(wiki), name(Name)]
+	;   Attrs = class(wiki),
+	    strip_ws_tokens(List, Content)
+	).
 
 
 %%	strip_ws_tokens(+Tokens, -Stripped)
@@ -285,7 +328,8 @@ strip_leading_ws(T, T).
 tags(Lines, Tags) :-
 	collect_tags(Lines, Tags0),
 	keysort(Tags0, Tags1),
-	combine_tags(Tags1, Tags).
+	pairs_values(Tags1, Tags2),
+	combine_tags(Tags2, Tags).
 
 %%	collect_tags(+IndentedLines, -Tags) is det
 %
@@ -341,44 +385,49 @@ renamed_tag(exception, throws).
 %	versions may issue a warning if the order is inconsistent.
 
 tag_order(param,       1).
-tag_order(throws,      2).
-tag_order(author,      3).
-tag_order(version,     4).
-tag_order(see,	       5).
-tag_order(deprecated,  6).
-tag_order(compat,      7).		% PlDoc extension
-tag_order(copyright,   8).
-tag_order(license,     9).
-tag_order(bug,	      10).
-tag_order(tbd,	      11).
-tag_order(error,       2).
+tag_order(error,       2).		% same as throw
+tag_order(throws,      3).
+tag_order(author,      4).
+tag_order(version,     5).
+tag_order(see,	       6).
+tag_order(deprecated,  7).
+tag_order(compat,      8).		% PlDoc extension
+tag_order(copyright,   9).
+tag_order(license,    10).
+tag_order(bug,	      11).
+tag_order(tbd,	      12).
 
 
-%%	combine_tags(+TaggedTags:list, -Tags:list) is det.
+%%	combine_tags(+Tags:list(tag(Key, Value)), -Tags:list) is det.
 %
 %	Creates the final tag-list.  Tags is a list of
 %	
-%		* \params(List of param(Name, Descr))
-%		* \tag(Name, Descr)
+%		* \params(list(param(Name, Descr)))
+%		* \tag(Name, list(Descr))
 %	
 %	Descr is a list of tokens.
 
 combine_tags([], []).
-combine_tags([_-tag(param, V1)|T0], [\params([P1|PL])|Tags]) :- !,
+combine_tags([tag(param, V1)|T0], [\params([P1|PL])|Tags]) :- !,
 	param_tag(V1, P1),
 	param_tags(T0, PL, T1),
 	combine_tags(T1, Tags).
-combine_tags([_-tag(Tag,Value)|T0], [\tag(Tag, Value)|T]) :-
-	combine_tags(T0, T).
+combine_tags([tag(Tag,V0)|T0], [\tag(Tag, [V0|Vs])|T]) :-
+	same_tag(Tag, T0, T1, Vs),
+	combine_tags(T1, T).
 
 param_tag([PN|Descr0], param(PN, Descr)) :-
 	strip_leading_ws(Descr0, Descr).
 
-param_tags([_-tag(param, V1)|T0], [P1|PL], T) :- !,
+param_tags([tag(param, V1)|T0], [P1|PL], T) :- !,
 	param_tag(V1, P1),
 	param_tags(T0, PL, T).
 param_tags(T, [], T).
 
+same_tag(Tag, [tag(Tag, V)|T0], T, [V|Vs]) :- !,
+	same_tag(Tag, T0, T, Vs).
+same_tag(_, L, L, []).
+
 
 		 /*******************************
 		 *	       FACES		*
@@ -387,7 +436,7 @@ param_tags(T, [], T).
 %%	wiki_faces(+Structure, +ArgNames, -HTML) is det.
 %
 %	Given the wiki structure, analyse the content of the paragraphs,
-%	list items and gtable cells and apply font faces and links.
+%	list items and table cells and apply font faces and links.
 
 wiki_faces(DOM0, ArgNames, DOM) :-
 	structure_term(DOM0, Functor, Content0), !,
@@ -404,10 +453,17 @@ wiki_faces_list([H0|T0], Args, [H|T]) :-
 	wiki_faces(H0, Args, H),
 	wiki_faces_list(T0, Args, T).
 
+%%	structure_term(+Term, -Functor, -Content) is semidet.
+%%	structure_term(-Term, +Functor, +Content) is det.
+%
+%	(Un)pack a term describing structure, so  we can process Content
+%	and re-pack the structure.
+
 structure_term(\tags(Tags), tags, [Tags]) :- !.
 structure_term(\params(Params), params, [Params]) :- !.
 structure_term(param(Name,Descr), param(Name), [Descr]) :- !.
 structure_term(\tag(Name,Value), tag(Name), [Value]) :- !.
+structure_term(\include(What,Type), include(What,Type), []) :- !.
 structure_term(dl(Att, Args), dl(Att), [Args]) :- !.
 structure_term(dt(Att, Args), dt(Att), [Args]) :- !.
 structure_term(table(Att, Args), table(Att), [Args]) :- !.
@@ -439,6 +495,10 @@ verbatim_term(pre(_,_)).
 verbatim_term(\term(_,_)).
 
 %%	wiki_faces(-WithFaces, +ArgNames)// is nondet.
+%
+%	Apply font-changes and automatic  links   to  running  text. The
+%	faces are applied after discovering   the structure (paragraphs,
+%	lists, tables, keywords).
 
 wiki_faces([], _) -->
 	[].
@@ -446,10 +506,6 @@ wiki_faces([H|T], ArgNames) -->
 	wiki_face(H, ArgNames),
 	wiki_faces(T, ArgNames).
 
-word_token(Word) -->
-	[Word],
-	{ string(Word) }.
-
 wiki_face(var(Word), ArgNames) -->
 	[Word],
 	{ string(Word),			% punctuation and blanks are atoms
@@ -469,28 +525,28 @@ wiki_face(code(Code), _) -->
 wiki_face(code(Code), _) -->
 	[=,'|'], wiki_faces(Code, []), ['|',=], !.
 wiki_face(\predref(Name/Arity), _) -->
-	[ NameS, '/', ArityWord ],
-	{ functor_name(NameS),
-	  catch(atom_number(ArityWord, Arity), _, fail),
-	  Arity >= 0, Arity < 100, !,
+	[ NameS, '/' ], arity(Arity),
+	{ functor_name(NameS), !,
 	  string_to_atom(NameS, Name)
 	}.
+wiki_face(\predref(Name/Arity), _) -->
+	symbol_string(S), [ '/' ], arity(Arity), !,
+	{ atom_chars(Name, S)
+	}.
 wiki_face(\predref(Name//Arity), _) -->
-	[ NameS, '/', '/', ArityWord ],
-	{ functor_name(NameS),
-	  catch(atom_number(ArityWord, Arity), _, fail),
-	  Arity >= 0, Arity < 100, !,
+	[ NameS, '/', '/' ], arity(Arity),
+	{ functor_name(NameS), !,
 	  string_to_atom(NameS, Name)
 	}.
 wiki_face(span(class=cvs, CVS), _) -->
 	[$, Word, :], {string(Word)}, wiki_faces(CVS0, []), [$], !,
 	{ strip_ws_tokens(CVS0, CVS) }.
-wiki_face(\image(Name), _) -->
+wiki_face(\include(Name, Type), _) -->
 	['[','['], word_token(BaseS), ['.'], word_token(ExtS), [']',']'],
 	{  concat_atom([BaseS, '.', ExtS], Name),
 	   file_name_extension(_, Ext, Name),
-	   autolink_extension(Ext, image)
-	}.
+	   autolink_extension(Ext, Type)
+	}, !.
 wiki_face(\file(Name), _) -->
 	word_token(BaseS), ['.'], word_token(ExtS),
 	{ concat_atom([BaseS, '.', ExtS], Name),
@@ -518,9 +574,68 @@ wiki_face(FT, ArgNames) -->
 	;   wiki_faces(T, ArgNames, FT)
 	}.
 
+%%	word_token(-Word:string)// is semidet.
+%
+%	True if the next token  is  a   string,  which  implies  it is a
+%	sequence of alpha-numerical characters.
+
+word_token(Word) -->
+	[Word],
+	{ string(Word) }.
+
+%%	arity(-Arity:int)// is semidet.
+%
+%	True if the next token can be  interpreted as an arity. That is,
+%	refers to a non-negative integers of at most 20. Although Prolog
+%	allows for higher arities, we assume 20   is  a fair maximum for
+%	user-created predicates that are documented.
+
+arity(Arity) -->
+	[ Word ],
+	{ catch(atom_number(Word, Arity), _, fail),
+	  Arity >= 0, Arity < 20
+	}.
+
+%%	symbol_string(-String)// is nondet
+%
+%	Accept  a  non-empty  sequence  of   Prolog  symbol  characters,
+%	starting with the shortest match.
+
+symbol_string([S]) -->
+	[S],
+	{ prolog_symbol_char(S) }.
+symbol_string([H|T]) -->
+	[H],
+	{ prolog_symbol_char(H) },
+	symbol_string(T).
+
+%%	prolog_symbol_char(?Char)
+%
+%	True if char is classified by Prolog as a symbol char.
+
+prolog_symbol_char(#).
+prolog_symbol_char($).
+prolog_symbol_char(&).
+prolog_symbol_char(*).
+prolog_symbol_char(+).
+prolog_symbol_char(-).
+prolog_symbol_char(.).
+prolog_symbol_char(/).
+prolog_symbol_char(:).
+prolog_symbol_char(<).
+prolog_symbol_char(=).
+prolog_symbol_char(>).
+prolog_symbol_char(?).
+prolog_symbol_char(@).
+prolog_symbol_char(\).
+prolog_symbol_char(^).
+prolog_symbol_char(`).
+prolog_symbol_char(~).
+
+
 functor_name(String) :-
 	sub_atom(String, 0, 1, _, Char),
-	char_type(Char, alpha).
+	char_type(Char, lower).
 
 url_protocol(String) :-	sub_atom(String, 0, _, 0, http).
 url_protocol(String) :-	sub_atom(String, 0, _, 0, ftp).
@@ -528,18 +643,22 @@ url_protocol(String) :-	sub_atom(String, 0, _, 0, mailto).
 
 
 peek_end_url -->
-	peek_nl, !.
-peek_end_url -->
-	peek(' '), !.
+	peek(End),
+	{ space_atom(End) }, !.
 peek_end_url -->
 	eos, !.
 
+space_atom(' ').
+space_atom('\r').
+space_atom('\n').
+
 %%	autolink_extension(?Ext, ?Type) is nondet.
 %
 %	True if Ext is a filename extensions that create automatic links
 %	in the documentation.
 
-autolink_extension(pl, prolog).
+autolink_extension(Ext, prolog) :-
+	prolog_file_type(Ext,prolog).
 autolink_extension(txt, wiki).
 autolink_extension(gif, image).
 autolink_extension(png, image).
@@ -914,14 +1033,6 @@ non_ws(T, T) -->
 	[].
 
 
-%%	peek_nl//
-%
-%	True if we are at the end of a line
-
-peek_nl -->
-	peek(H),
-	{ H == 0'\n ; H == 0'\r }, !.
-
 %%	nl//
 %
 %	Get end-of-line
