@@ -85,7 +85,7 @@ COMMON(void) 		do_undo(mark *m);
 COMMON(Definition) 	getProcDefinition(Procedure proc);
 COMMON(void) 		fix_term_ref_count(void);
 COMMON(fid_t) 		PL_open_signal_foreign_frame(void);
-COMMON(bool)		foreignWakeup(ARG1_LD, term_t *ex);
+COMMON(bool)		foreignWakeup(term_t *ex ARG_LD);
 
 /* pl-atom.c */
 COMMON(word) 		lookupAtom(const char *s, size_t len);
@@ -448,7 +448,6 @@ COMMON(void) 		resetForeign(void);
 
 /* pl-list.c */
 COMMON(word) 		pl_length(term_t list, term_t l);
-COMMON(word) 		pl_memberchk(term_t e, term_t list);
 
 /* pl-load.c */
 COMMON(word) 		pl_open_shared_object(term_t file, term_t h, term_t flags);
@@ -541,6 +540,7 @@ COMMON(int) 		compareStandard(Word t1, Word t2, int eq ARG_LD);
 COMMON(intptr_t)	skip_list(Word l, Word *tailp ARG_LD);
 COMMON(intptr_t) 	lengthList(term_t list, int errors);
 COMMON(int) 		numberVars(term_t t, nv_options *opts, int n ARG_LD);
+COMMON(int)		duplicate_term(term_t in, term_t copy ARG_LD);
 COMMON(word) 		pl_e_free_variables(term_t t, term_t l);
 COMMON(word) 		stringToList(char *s);
 COMMON(word) 		pl_atom_concat(term_t a1, term_t a2, term_t a3,
@@ -686,7 +686,6 @@ COMMON(word) 		pl_recordz(term_t key, term_t term, term_t ref);
 COMMON(word) 		pl_recorded(term_t key, term_t term, term_t ref, control_t h);
 COMMON(word) 		pl_erase(term_t ref);
 COMMON(word) 		pl_term_complexity(term_t t, term_t mx, term_t count);
-COMMON(void) 		undo_while_saving_term(mark *m, Word term);
 COMMON(void)		markAtomsRecord(Record record);
 
 /* pl-rl.c */
