@@ -1,11 +1,11 @@
-/*  $Id: pl-gc.c,v 1.128 2007/06/01 20:30:02 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -119,15 +119,17 @@ char tmp[256];				/* for calling print_val(), etc. */
 #define recordMark(p)
 #define needsRelocation(p) { needs_relocation++; }
 #define check_relocation(p)
+#define markLocal(p) (local_marked++)
+#define processLocal(p) (local_marked--)
 #endif
 
 #define ldomark(p)	{ *(p) |= MARK_MASK; }
 #define domark(p)	{ if ( is_marked(p) ) \
 			    sysError("marked twice: %p (*= 0x%lx), gTop = %p", p, *(p), gTop); \
+			  DEBUG(3, char b[64]; Sdprintf("\tdomarked(%p = %s)\n", p, print_val(*p, b))); \
 			  *(p) |= MARK_MASK; \
 			  total_marked++; \
 			  recordMark(p); \
-			  DEBUG(4, Sdprintf("is_marked(%p)\n", p)); \
 			}
 #define unmark(p)	(*(p) &= ~MARK_MASK)
 
@@ -152,7 +154,7 @@ char tmp[256];				/* for calling print_val(), etc. */
 #define onTrail(p)	topPointerOnStack(trail, p)
 
 #ifndef offset
-#define offset(s, f) ((int)(&((struct s *)NULL)->f))
+#define offset(s, f) ((size_t)(&((struct s *)NULL)->f))
 #endif
 
 #define ttag(x)		(((word)(x))&TAG_TRAILMASK)
@@ -164,7 +166,9 @@ char tmp[256];				/* for calling print_val(), etc. */
 forwards void		mark_variable(Word ARG_LD);
 forwards void		sweep_foreign(void);
 static void		sweep_global_mark(Word *m ARG_LD);
+#ifndef LIFE_GC
 forwards QueryFrame	mark_environments(LocalFrame, Code PC);
+#endif
 forwards void		update_relocation_chain(Word, Word ARG_LD);
 forwards void		into_relocation_chain(Word, int stg ARG_LD);
 forwards void		alien_into_relocation_chain(void *addr,
@@ -207,6 +211,7 @@ static int		check_marked(const char *s);
 #define mark_base	   (LD->gc._mark_base)
 #define mark_top	   (LD->gc._mark_top)
 #define check_table	   (LD->gc._check_table)
+#define local_table	   (LD->gc._local_table)
 #endif
 
 #undef LD
@@ -254,18 +259,39 @@ print_adr(Word adr, char *buf)
 
 static char *
 print_val(word val, char *buf)
-{ char *tag_name[] = { "var", "float", "int", "atom",
-		       "string", "list", "term", "ref" };
-  char *stg_name[] = { "static/inline/trail", "global", "local", "reserved" };
+{ GET_LD
+  char *tag_name[] = { "var", "attvar", "float", "int", "atom",
+		       "string", "term", "ref" };
+  char *stg_name[] = { "static", "global", "local", "reserved" };
+  char *o = buf;
 
-  Ssprintf(buf, "%s at %s(%ld)",
-	   tag_name[tag(val)],
-	   stg_name[storage(val) >> 3],
-	   (val >> LMASK_BITS)/sizeof(word));
-  if ( val & MARK_MASK )
-    strcat(buf, "M");
-  if ( val & FIRST_MASK )
-    strcat(buf, "F");
+  if ( val & (MARK_MASK|FIRST_MASK) )
+  { *o++ = '[';
+    if ( val & MARK_MASK )
+      *o++ = 'M';
+    if ( val & FIRST_MASK )
+      *o++ = 'F';
+    *o++ = ']';
+    val &= ~(word)(MARK_MASK|FIRST_MASK);
+  }
+
+  if ( isVar(val) )
+    strcpy(o, "VAR");
+  else if ( isTaggedInt(val) )
+    Ssprintf(o, "int(%ld)", valInteger(val));
+  else if ( isAtom(val) )
+  { const char *s = stringAtom(val);
+    if ( strlen(s) > 10 )
+    { strncpy(o, s, 10);
+      strcat(o, "...");
+    } else
+    { strcpy(o, s);
+    }
+  } else
+    Ssprintf(o, "%s at %s(%ld)",
+	     tag_name[tag(val)],
+	     stg_name[storage(val) >> 3],
+	     (val >> LMASK_BITS)/sizeof(word));
 
   return buf;
 }
@@ -290,6 +316,28 @@ do_check_relocation(Word addr, char *file, int line ARG_LD)
   s->value = FALSE;
 }
 
+
+static void
+markLocal(Word addr)
+{ GET_LD
+  
+  local_marked++;
+  addHTable(local_table, addr, (Void)TRUE);
+}
+
+static void
+processLocal(Word addr)
+{ GET_LD
+  Symbol s;
+
+  local_marked--;
+  if ( (s = lookupHTable(local_table, addr)) )
+  { s->value = (Void)FALSE;
+  } else
+  { assert(0);
+  }
+}
+
 #endif /* O_SECURE */
 
 		/********************************
@@ -359,7 +407,7 @@ unmark_environments(PL_local_data_t *ld, LocalFrame fr, uintptr_t mask)
     if ( fr->parent )
       fr = fr->parent;
     else				/* Prolog --> C --> Prolog calls */
-      return (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+      return QueryOfTopFrame(fr);
   }
 }
 
@@ -427,16 +475,20 @@ mark_variable(Word start ARG_LD)
   word val;				/* old value of current cell */
   Word next;				/* cell to be examined */
 
-  DEBUG(3, Sdprintf("marking %p\n", start));
+  DEBUG(3,
+	char b[64];
+	Sdprintf("marking %p (=%s)\n", start, print_val(*start, b)));
 
   if ( is_marked(start) )
     sysError("Attempt to mark twice");
 
-  local_marked++;
+  if ( onStackArea(local, start) )
+  { markLocal(start);
+    total_marked--;			/* do not count local stack cell */
+  }
   current = start;
   mark_first(current);
   val = get_value(current);  
-  total_marked--;			/* do not count local stack cell */
   FORWARD;
 
 forward:				/* Go into the tree */
@@ -568,6 +620,12 @@ static void
 mark_term_refs()
 { GET_LD
   FliFrame fr = fli_context;
+#if O_DEBUG
+  long gmarked = 0;
+  long lmarked = 0;
+#endif
+
+  DEBUG(3, Sdprintf("Marking term references ...\n"));
 
   for( ; fr; fr = fr->parent )
   { Word sp = refFliP(fr, 0);
@@ -578,14 +636,19 @@ mark_term_refs()
     { SECURE(assert(!is_marked(sp)));
 
       if ( isGlobalRef(*sp) )
-      { mark_variable(sp PASS_LD);
+      { DEBUG(3, gmarked++);
+	mark_variable(sp PASS_LD);
       } else
-      { ldomark(sp);      
+      { DEBUG(3, lmarked++);
+	ldomark(sp);      
       }
     }
 
     SECURE(check_marked("After marking foreign frame"));
   }
+
+  DEBUG(3, Sdprintf("Marked %ld global and %ld local term references\n",
+		    gmarked, lmarked));
 }
 
 
@@ -613,7 +676,6 @@ static fid_t
 gvars_to_term_refs(Word **saved_bar_at)
 { GET_LD
   fid_t fid;
-  Word *sb;
 
   *saved_bar_at = NULL;
 
@@ -637,13 +699,16 @@ gvars_to_term_refs(Word **saved_bar_at)
     freeTableEnum(e);
     assert(LD->gvar.grefs == found);
 
-    DEBUG(0, Sdprintf("Found %d global vars on global stack. "
+    DEBUG(1, Sdprintf("Found %d global vars on global stack. "
 		      "stored in frame %p\n", found, fli_context));
   } else
     fid = 0;
 
   if ( LD->frozen_bar )
-  { sb = (Word*)lTop;
+  { Word *sb;
+
+    requireStack(local, sizeof(Word));
+    sb = (Word*)lTop;
     lTop = (LocalFrame)(sb+1);
     *sb = LD->frozen_bar;
     *saved_bar_at = sb;
@@ -695,6 +760,22 @@ term_refs_to_gvars(fid_t fid, Word *saved_bar_at)
 #endif /*O_GVAR*/
 
 
+#ifdef O_CALL_RESIDUE
+static void
+mark_attvars()
+{ GET_LD
+  Word gp;
+
+  for( gp = gBase; gp < gTop; gp += (offset_cell(gp)+1) )
+  { if ( isAttVar(*gp) && !is_marked(gp) )
+    { DEBUG(3, Sdprintf("mark_attvars(): marking %p\n", gp));
+      mark_variable(gp PASS_LD);
+    }
+  }
+}
+#endif /*O_CALL_RESIDUE*/
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 clearUninitialisedVarsFrame(LocalFrame fr, Code PC);
 
@@ -730,6 +811,7 @@ clearUninitialisedVarsFrame(LocalFrame fr, Code PC)
 	case B_MPZ:
 	{ word m = PC[1];
 	  PC += wsizeofInd(m)+1;
+	  PC -= VM_DYNARGC;		/* compensate for codeTable[c].arguments */
 	  break;
 	}
 	case I_EXIT:
@@ -768,6 +850,22 @@ slotsInFrame(LocalFrame fr, Code PC)
 }
 
 
+static inline void
+check_call_residue(LocalFrame fr ARG_LD)
+{
+#ifdef O_CALL_RESIDUE
+  if ( fr->predicate == PROCEDURE_call_residue_vars2->definition )
+  { if ( !LD->gc.marked_attvars )
+    { mark_attvars();
+      LD->gc.marked_attvars = TRUE;
+    }
+  }
+#endif
+}
+
+
+#ifndef LIFE_GC
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Mark environments and all data that can   be  reached from them. Returns
 the QueryFrame that started this environment,  which provides use access
@@ -795,6 +893,7 @@ mark_environments(LocalFrame fr, Code PC)
 		      levelFrame(fr), predicateName(fr->predicate)));
 
     clearUninitialisedVarsFrame(fr, PC);
+    check_call_residue(fr PASS_LD);
 
     slots  = slotsInFrame(fr, PC);
 #if O_SECURE
@@ -814,10 +913,11 @@ mark_environments(LocalFrame fr, Code PC)
     if ( fr->parent != NULL )
       fr = fr->parent;
     else
-      return (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+      return QueryOfTopFrame(fr);
   }
 }
 
+#endif /*not LIFE_GC*/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 If multiple TrailAssignment() calls happen on  the same address within a
@@ -864,15 +964,6 @@ mergeTrailedAssignments(GCTrailEntry top, GCTrailEntry mark,
 
   DEBUG(2, Sdprintf("Scanning %d trailed assignments\n", assignments));
 
-#if O_SECURE
-  for(te=mark; te <= top; te++)
-  { if ( ttag(te[1].address) == TAG_TRAILVAL )
-    { Word p = val_ptr(te->address);
-      assert(!is_first(p));
-    }
-  }
-#endif
-
   for(te=mark; te <= top; te++)
   { if ( ttag(te[1].address) == TAG_TRAILVAL )
     { Word p = val_ptr(te->address);
@@ -908,12 +999,8 @@ normal trail-pointer, while the  second   is  flagged  with TAG_TRAILVAL
 is encountered, this value is restored  at   the  location  of the first
 trail-cell.
 
-If the trail cell has become garbage,   we  can destroy both cells. Note
-that mark_trail() already has marked the   replaced  value, so that will
-not be garbage collected during this pass. We cannot reverse marking the
-stacks and the trail-stack as one trailed   value might make another one
-non-garbage. It isn't too bad however as the   next GC will take care of
-the term.
+If the trail cell  has  become  garbage,   we  can  destroy  both cells,
+otherwise we must mark the value.
 
 Early reset of trailed  assignments  is   another  issue.  If  a trailed
 location has not yet been  marked  it   can  only  be accessed by frames
@@ -943,12 +1030,34 @@ early_reset_vars(mark *m, Word top, GCTrailEntry te ARG_LD)
 	te->address = 0;
 	trailcells_deleted += 2;
       } else if ( is_marked(tard) )
-      { assignments++;
+      { Word gp = val_ptr(te->address);
+
+	assert(onGlobal(gp));
+	assert(!is_first(gp));
+	if ( !is_marked(gp) )
+	{ DEBUG(3,
+		char b1[64]; char b2[64]; char b3[64];
+		Sdprintf("Marking assignment at %s (%s --> %s)\n",
+			 print_adr(tard, b1),
+			 print_val(*gp, b2),
+			 print_val(*tard, b3)));
+
+	  mark_variable(gp PASS_LD);
+	  assert(is_marked(gp));
+	}
+
+	assignments++;
+	te--;
       } else
-      { Word gp;
-	DEBUG(3, Sdprintf("Early reset of assignment at %p (*=0x%lx)\n",
-		 tard, *tard));
-	gp = val_ptr(te->address);
+      { Word gp = val_ptr(te->address);
+
+	DEBUG(4,
+	      char b1[64]; char b2[64]; char b3[64];
+	      Sdprintf("Early reset of assignment at %s (%s --> %s)\n",
+		       print_adr(tard, b1),
+		       print_val(*tard, b2),
+		       print_val(*gp, b3)));
+
 	assert(onGlobal(gp));
 	*tard = *gp;
 	unmark(tard);
@@ -966,9 +1075,11 @@ early_reset_vars(mark *m, Word top, GCTrailEntry te ARG_LD)
       { SECURE(assert(ttag(te[1].address) != TAG_TRAILVAL));
 	te->address = 0;
 	trailcells_deleted++;
-      } else if ( !is_marked(tard) )	/* garbage */
-      { setVar(*tard);
-	DEBUG(3, Sdprintf("Early reset at %p\n", tard));
+      } else if ( !is_marked(tard) )
+      { DEBUG(3,
+	      char b1[64]; char b2[64]; 
+	      Sdprintf("Early reset at %s (%s)\n", print_adr(tard, b1), print_val(*tard, b2)));
+	setVar(*tard);
 	te->address = 0;
 	trailcells_deleted++;
       }
@@ -1001,6 +1112,7 @@ mark_foreign_frame(FliFrame fr, GCTrailEntry te)
 }
 
 
+#ifndef LIFE_GC
 static GCTrailEntry
 mark_choicepoints(Choice ch, GCTrailEntry te, FliFrame *flictx)
 { GET_LD
@@ -1029,12 +1141,12 @@ mark_choicepoints(Choice ch, GCTrailEntry te, FliFrame *flictx)
     alien_into_relocation_chain(&ch->mark.trailtop,
 				STG_TRAIL, STG_LOCAL PASS_LD);
     SECURE(trailtops_marked--);
-
     mark_environments(fr, ch->type == CHP_JUMP ? ch->value.PC : NULL);
   }
 
   return te;
 }
+#endif
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1046,58 +1158,49 @@ relocation chains. A small problem is  the   top-goal  of  a query, This
 frame may not be a  choicepoint,  but   its  mark  is  needed anyhow for
 PL_close_query(), so it has to be relocated.  `te' in the function below
 has to be updated as none of these variables should be reset
+
+We must first mark all environments,   including  those in outer queries
+(Prolog -> C -> Prolog calls) as mark_choicepoints() can otherwise reset
+variables in outer queries.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#ifdef LIFE_GC
+#include "pl-lifegc.c"
+#else
 static void
 mark_stacks(LocalFrame fr, Choice ch)
 { GET_LD
-  QueryFrame query;
+  QueryFrame qf=NULL, pqf=NULL, top = NULL;
   GCTrailEntry te = (GCTrailEntry)tTop - 1;
   FliFrame flictx = fli_context;
 
   trailcells_deleted = 0;
 
-  for( ; fr; fr = query->saved_environment, ch = query->saved_bfr )
-  { query = mark_environments(fr, NULL);
-    te    = mark_choicepoints(ch, te, &flictx);
+  for( ; fr; fr = qf->saved_environment )
+  { qf = mark_environments(fr, NULL);
 
-    assert(query->magic == QID_MAGIC);
+    assert(qf->magic == QID_MAGIC);
+
+    if ( pqf )
+    { pqf->parent = qf;
+    } else if ( !top )
+    { top = qf;
+    } 
+    pqf = qf;
   }
+  if ( qf )
+    qf->parent = NULL;			/* topmost query */
   
+  te = mark_choicepoints(ch, te, &flictx);
+  for(qf=top; qf; qf=qf->parent)
+    te = mark_choicepoints(qf->saved_bfr, te, &flictx);
+
   for( ; flictx; flictx = flictx->parent)
     te = mark_foreign_frame(flictx, te);
 
   DEBUG(2, Sdprintf("Trail stack garbage: %ld cells\n", trailcells_deleted));
 }
-
-
-#ifdef O_DESTRUCTIVE_ASSIGNMENT
-static void
-mark_trail()
-{ GET_LD
-  GCTrailEntry te = (GCTrailEntry)tTop - 1;
-
-  for( ; te >= (GCTrailEntry)tBase; te-- )
-  { Word gp;
-
-    if ( ttag(te->address) == TAG_TRAILVAL )
-    { gp = val_ptr(te->address);
-
-      DEBUG(3, Sdprintf("mark_trail(): trailed value from %p at %p (*=%p)\n",
-			&te->address, gp, *gp));
-
-      assert(onGlobal(gp));
-      if ( !is_marked(gp) )
-      {	total_marked++;			/* fix counters */
-	local_marked--;
-
-	mark_variable(gp PASS_LD);
-	assert(is_marked(gp));
-      }
-    }
-  }
-}
-#endif /*O_DESTRUCTIVE_ASSIGNMENT*/
+#endif
 
 
 #if O_SECURE
@@ -1118,9 +1221,6 @@ mark_phase(LocalFrame fr, Choice ch)
 
   SECURE(check_marked("Before mark_term_refs()"));
   mark_term_refs();
-#ifdef O_DESTRUCTIVE_ASSIGNMENT
-  mark_trail();
-#endif
   mark_stacks(fr, ch);
 #if O_SECURE
   if ( !scan_global(TRUE) )
@@ -1171,7 +1271,7 @@ update_relocation_chain(Word current, Word dest ARG_LD)
 { Word head = current;
   word val = get_value(current);
 
-  DEBUG(3, Sdprintf("unwinding relocation chain at %p to %p\n",
+  DEBUG(4, Sdprintf("unwinding relocation chain at %p to %p\n",
 		    current, dest));
 
   do
@@ -1206,7 +1306,7 @@ into_relocation_chain(Word current, int stg ARG_LD)
   set_value(current, get_value(head));
   set_value(head, consPtr(current, stg|tag(val)));
 
-  DEBUG(2, Sdprintf("Into relocation chain: %p (head = %p)\n",
+  DEBUG(4, Sdprintf("Into relocation chain: %p (head = %p)\n",
 		    current, head));
 
   if ( is_first(head) )
@@ -1290,17 +1390,17 @@ tag_trail()
 
       SECURE(assert(onStack(global, p2)));
       te->address = (Word)consPtr(p2, STG_GLOBAL|TAG_TRAILVAL);
-      te--;
-    }
-
-    if ( onLocal(te->address) )
-    { stg = STG_LOCAL;
+      //SECURE(assert(te == tBase || !isTrailVal(te[-1].address)));
     } else
-    { SECURE(assert(onStackArea(global, te->address)));
-      stg = STG_GLOBAL;
+    { if ( onLocal(te->address) )
+      { stg = STG_LOCAL;
+      } else
+      { SECURE(assert(onStackArea(global, te->address)));
+	stg = STG_GLOBAL;
+      }
+	
+      te->address = (Word)consPtr(te->address, stg);
     }
-
-    te->address = (Word)consPtr(te->address, stg);
   }
 }
 
@@ -1379,7 +1479,7 @@ sweep_foreign()
     { if ( is_marked(sp) )
       {	unmark(sp);
 	if ( isGlobalRef(get_value(sp)) )
-	{ local_marked--;
+	{ processLocal(sp);
 	  check_relocation(sp);
 	  into_relocation_chain(sp, STG_LOCAL PASS_LD);
 	}
@@ -1421,7 +1521,7 @@ unsweep_environments(LocalFrame fr)
 { while(fr->parent)
     fr = fr->parent;
 
-  return (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+  return QueryOfTopFrame(fr);
 }
 
 
@@ -1486,18 +1586,30 @@ sweep_environments(LocalFrame fr, Code PC)
     { if ( is_marked(sp) )
       { unmark(sp);
 	if ( isGlobalRef(get_value(sp)) )
-	{ local_marked--;
+	{ processLocal(sp);
 	  check_relocation(sp);
 	  into_relocation_chain(sp, STG_LOCAL PASS_LD);
 	}
       }
+#ifdef LIFE_GC
+      else
+      { if ( isGlobalRef(*sp) )
+	{ DEBUG(1, char b[64];
+		Sdprintf("[%ld] %s: GC VAR(%d) (=%s)\n",
+			 levelFrame(fr), predicateName(fr->predicate),
+			 sp-argFrameP(fr, 0),
+			 print_val(*sp, b)));
+	  *sp = ATOM_garbage_collected;
+	}
+      }
+#endif
     }
 
     PC = fr->programPointer;
     if ( fr->parent != NULL )
       fr = fr->parent;
     else
-      return (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+      return QueryOfTopFrame(fr);
   }
 }
 
@@ -1526,7 +1638,27 @@ sweep_stacks(LocalFrame fr, Choice ch)
   }
 
   if ( local_marked != 0 )
+  {
+#ifdef O_SECURE
+    TableEnum e = newTableEnum(local_table);
+    Symbol s;
+
+    Sdprintf("FATAL: unprocessed local variables:\n");
+
+    while((s=advanceTableEnum(e)))
+    { if ( s->value )
+      {	Word p = s->name;
+	char buf1[64];
+	char buf2[64];
+
+	Sdprintf("\t%s (*= %s)\n", print_adr(p, buf1), print_val(*p, buf2));
+      }
+    }
+
+    freeTableEnum(e);
+#endif
     sysError("local_marked = %ld", local_marked);
+  }
 }
 
 
@@ -1647,7 +1779,7 @@ compact_global(void)
 		 current, *current, gTop, *v);
 #endif
       dest -= offset + 1;
-      DEBUG(3, Sdprintf("Marked cell at %p (size = %ld; dest = %p)\n",
+      DEBUG(4, Sdprintf("Marked cell at %p (size = %ld; dest = %p)\n",
 			current, offset+1, dest));
       if ( is_first(current) )
 	update_relocation_chain(current, dest PASS_LD);
@@ -1763,27 +1895,50 @@ collect_phase(LocalFrame fr, Choice ch, Word *saved_bar_at)
 		*********************************/
 
 #if O_DYNAMIC_STACKS
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+If s == NULL, consider all stacks
+
+(*) Do not  consider  GC  if  there   are  no  inferences.  This  avoids
+repetetive GC calls while building large   structures  from foreign code
+that calls PL_handle_signals() from time to   time  to enable interrupts
+and call GC.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 void
 considerGarbageCollect(Stack s)
 { GET_LD
-  if ( s->gc && trueFeature(GC_FEATURE) && !gc_status.requested )
-  { intptr_t used  = (char *)s->top   - (char *)s->base;
-    intptr_t free  = (char *)s->limit - (char *)s->top;
-    intptr_t limit = (char *)s->limit - (char *)s->base;
 
-    if ( used > s->factor*s->gced_size + s->small )
-    { DEBUG(2, Sdprintf("GC: request on %s, factor=%d, last=%ld, small=%ld\n",
-			s->name, s->factor, s->gced_size, s->small));
-      gc_status.requested = TRUE;
-    } else if ( free < limit/8 && used > s->gced_size + limit/32 ) 
-    { DEBUG(2, Sdprintf("GC: request on low free\n"));
-      gc_status.requested = TRUE;
+  if ( trueFeature(GC_FEATURE) && !PL_pending(SIG_GC) )
+  { if ( s == NULL )
+    { considerGarbageCollect((Stack)&LD->stacks.global);
+      considerGarbageCollect((Stack)&LD->stacks.trail);
+    } else
+    { if ( s->gc )
+      { intptr_t used  = (char *)s->top   - (char *)s->base;
+	intptr_t free  = (char *)s->limit - (char *)s->top;
+	intptr_t limit = (char *)s->limit - (char *)s->base;
+    
+	if ( LD->gc.inferences == LD->statistics.inferences )
+	{ s->gced_size = used;		/* (*) */
+	  return;
+	}
+    
+	if ( used > s->factor*s->gced_size + s->small )
+	{ DEBUG(2, Sdprintf("GC: request on %s, factor=%d, last=%ld, small=%ld\n",
+			    s->name, s->factor, s->gced_size, s->small));
+	  PL_raise(SIG_GC);
+	} else if ( free < limit/8 && used > s->gced_size + limit/32 ) 
+	{ DEBUG(2, Sdprintf("GC: request on low free\n"));
+	  PL_raise(SIG_GC);
+	}
+    
+	DEBUG(1, if ( PL_pending(SIG_GC) )
+		 { Sdprintf("%s overflow: Posted garbage collect request\n",
+			    s->name);
+		 });
+      }
     }
-
-    DEBUG(1, if ( gc_status.requested)
-	     { Sdprintf("%s overflow: Posted garbage collect request\n",
-			s->name);
-	     });
   }
 }
 #endif /* O_DYNAMIC_STACKS */
@@ -1879,9 +2034,12 @@ check_environments(LocalFrame fr, Code PC, Word key)
 
     assert(onStack(local, fr));
 
-    DEBUG(3, Sdprintf("Check [%ld] %s:",
+    DEBUG(3, Sdprintf("Check [%ld] %s (PC=%d):",
 		      levelFrame(fr),
-		      predicateName(fr->predicate)));
+		      predicateName(fr->predicate),
+		      (false(fr->predicate, FOREIGN) && PC)
+		        ? (PC-fr->clause->clause->codes)
+		      	: 0));
 
     slots = slotsInFrame(fr, PC);
     sp = argFrameP(fr, 0);
@@ -1894,7 +2052,7 @@ check_environments(LocalFrame fr, Code PC, Word key)
     if ( fr->parent )
       fr = fr->parent;
     else
-    { QueryFrame qf = (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+    { QueryFrame qf = QueryOfTopFrame(fr);
       DEBUG(3, Sdprintf("*** Query %s\n", predicateName(qf->frame.predicate)));
       return qf;
     }
@@ -1964,10 +2122,21 @@ check_trail()
 
       assert(onGlobal(gp));
       key += checkData(gp);
+      assert(te > tBase);
+      te--;
+      assert(!isTrailVal(te->address));
 #ifdef O_SECURE
     } else
     { if ( onStackArea(global, te->address) )
-	assert(onStack(global, te->address));
+      { if ( !onStack(global, te->address) )
+	{ char b1[64], b2[64], b3[64];
+
+	  Sdprintf("Trail entry at %s not on global stack: %s (*=%s)\n",
+		   print_adr(te, b1),
+		   print_adr(te->address, b2),
+		   print_val(*te->address, b3));
+	}
+      }
 #endif
     }
   }
@@ -2087,19 +2256,29 @@ garbageCollect(LocalFrame fr, Choice ch)
   sigset_t mask;
   fid_t fid;
   Word *saved_bar_at;
+#ifdef O_PROFILE
+  struct call_node *prof_node = NULL;
+#endif
 #ifdef O_SECURE
   word key;
 #endif
 
-  DEBUG(0, verbose = TRUE);
+  DEBUG(1, verbose = TRUE);
 
   if ( gc_status.blocked || !trueFeature(GC_FEATURE) )
     return;
 
+  if ( !fr )
+    fr = LD->environment;
+  if ( !ch )
+    ch = LD->choicepoints;
+
   enterGC();
+#ifndef UNBLOCKED_GC
   blockSignals(&mask);
+#endif
   blockGC(PASS_LD1);			/* avoid recursion due to */
-  gc_status.requested = FALSE;		/* printMessage() */
+  PL_clearsig(SIG_GC);
 
   gc_status.active = TRUE;
   if ( verbose )
@@ -2109,7 +2288,7 @@ garbageCollect(LocalFrame fr, Choice ch)
 
 #ifdef O_PROFILE
   if ( LD->profile.active )
-    profCall(PROCEDURE_garbage_collect0->definition PASS_LD);
+    prof_node = profCall(PROCEDURE_garbage_collect0->definition PASS_LD);
 #endif
 
 #if O_SECURE
@@ -2119,9 +2298,12 @@ garbageCollect(LocalFrame fr, Choice ch)
   key = checkStacks(fr, ch);
 
   if ( check_table == NULL )
-    check_table = newHTable(256);
-  else
-    clearHTable(check_table);
+  { check_table = newHTable(256);
+    local_table = newHTable(256);
+  } else
+  { clearHTable(check_table);
+    clearHTable(local_table);
+  }
 
   mark_base = mark_top = malloc(usedStack(global));
 #endif
@@ -2131,6 +2313,7 @@ garbageCollect(LocalFrame fr, Choice ch)
   relocation_cells  = 0;
   relocated_cells   = 0;
   local_marked	    = 0;
+  LD->gc.marked_attvars = FALSE;
 
   requireStack(global, sizeof(word));
   requireStack(trail, sizeof(struct trail_entry));
@@ -2142,8 +2325,8 @@ garbageCollect(LocalFrame fr, Choice ch)
   mark_phase(fr, ch);
   tgar = trailcells_deleted * sizeof(struct trail_entry);
   ggar = (gTop - gBase - total_marked) * sizeof(word);
-  gc_status.trail_gained  += tgar;
   gc_status.global_gained += ggar;
+  gc_status.trail_gained  += tgar;
   gc_status.collections++;
 
   DEBUG(2, Sdprintf("Compacting trail ... "));
@@ -2164,12 +2347,18 @@ garbageCollect(LocalFrame fr, Choice ch)
   trimStacks(PASS_LD1);
   LD->stacks.global.gced_size = usedStack(global);
   LD->stacks.trail.gced_size  = usedStack(trail);
+  gc_status.global_left      += usedStack(global);
+  gc_status.trail_left       += usedStack(trail);
   gc_status.active = FALSE;
 
+#if LIFE_GC == 1
+  SECURE(checkStacks(fr, ch));
+#else
   SECURE(if ( checkStacks(fr, ch) != key )
 	 { sysError("ERROR: Stack checksum failure\n");
 	   trap_gdb();
 	 });
+#endif
 
   if ( verbose )
     printMessage(ATOM_informational,
@@ -2184,21 +2373,21 @@ garbageCollect(LocalFrame fr, Choice ch)
 		     PL_INTPTR, roomStack(trail));
 
 #ifdef O_PROFILE
-  if ( fr && LD->profile.active )
-    profExit(fr->prof_node PASS_LD);
+  if ( prof_node && LD->profile.active )
+    profExit(prof_node PASS_LD);
 #endif
 
   unblockGC(PASS_LD1);
+#ifndef UNBLOCKED_GC
   unblockSignals(&mask);
+#endif
+  LD->gc.inferences = LD->statistics.inferences;
   leaveGC();
 }
 
 word
 pl_garbage_collect(term_t d)
-{ GET_LD
-  LocalFrame fr = environment_frame;
-  Choice ch = LD->choicepoints;
-
+{ 
 #if O_DEBUG
   int ol = GD->debug_level;
   int nl;
@@ -2209,7 +2398,7 @@ pl_garbage_collect(term_t d)
     GD->debug_level = nl;
   }
 #endif
-  garbageCollect(fr, ch);
+  garbageCollect(NULL, NULL);
 #if O_DEBUG
   GD->debug_level = ol;
 #endif
@@ -2320,7 +2509,7 @@ update_environments(LocalFrame fr, Code PC, intptr_t ls, intptr_t gs, intptr_t t
 	update_pointer(&fr->clause->clause->codes, ls);
       }
 
-					/* update saved BFR's from C_MARK */
+					/* update saved BFR's from C_IFTHEN */
       if ( PC && false(fr->predicate, FOREIGN) )
       { Clause cl = fr->clause->clause;
 	unsigned int marks;
@@ -2343,7 +2532,7 @@ update_environments(LocalFrame fr, Code PC, intptr_t ls, intptr_t gs, intptr_t t
     if ( fr->parent )
       fr = fr->parent;
     else				/* Prolog --> C --> Prolog calls */
-    { QueryFrame query = (QueryFrame)addPointer(fr, -offset(queryFrame,frame));
+    { QueryFrame query = QueryOfTopFrame(fr);
 
       if ( ls )
       { update_pointer(&query->saved_bfr, ls);
@@ -2670,7 +2859,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC,
   enterGC();				/* atom-gc synchronisation */
   blockSignals(&mask);
   blockGC(PASS_LD1);			/* avoid recursion due to */
-  gc_status.requested = FALSE;		/* printMessage() */
+  PL_clearsig(SIG_GC);
 
   if ( !fr )
     fr = environment_frame;
@@ -2683,7 +2872,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC,
     double time = CpuTime(CPU_USER);
     int verbose = trueFeature(TRACE_GC_FEATURE);
     
-    DEBUG(0, verbose = TRUE);
+    DEBUG(1, verbose = TRUE);
 
     if ( verbose )
     { printMessage(ATOM_informational,
@@ -2750,7 +2939,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC,
 	}
 
 
-    DEBUG(0, { Sputchar('\n');
+    DEBUG(1, { Sputchar('\n');
 	       PrintStackParms(global, "global", gb, gsize);
 	       PrintStackParms(local, "local", lb, lsize);
 	       PrintStackParms(trail, "trail", tb, tsize);
@@ -2880,6 +3069,9 @@ mark_atoms_in_environments(PL_local_data_t *ld, LocalFrame fr)
     ld->gc._local_frames++;
     clearUninitialisedVarsFrame(fr, PC);
 
+    if ( fr->predicate == PROCEDURE_dcall1->definition )
+      forAtomsInClause(fr->clause->clause, markAtom);
+
     if ( true(fr->predicate, FOREIGN) ||
 	 !fr->clause )
       slots = fr->predicate->functor->arity;
@@ -2896,9 +3088,7 @@ mark_atoms_in_environments(PL_local_data_t *ld, LocalFrame fr)
     if ( fr->parent )
       fr = fr->parent;
     else
-    { QueryFrame qf = (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
-      return qf;
-    }
+     return QueryOfTopFrame(fr);
   }
 }
 
@@ -3079,9 +3269,7 @@ mark_predicates_in_environments(PL_local_data_t *ld, LocalFrame fr)
     if ( fr->parent )
       fr = fr->parent;
     else
-    { QueryFrame qf = (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
-      return qf;
-    }
+      return QueryOfTopFrame(fr);
   }
 }
 
@@ -3120,4 +3308,7 @@ BeginPredDefs(gc)
 #if O_SECURE || O_DEBUG || defined(O_MAINTENANCE)
   PRED_DEF("$check_stacks", 1, check_stacks, 0)
 #endif
+#ifdef GC_COUNTING
+  PRED_DEF("gc_statistics", 1, gc_statistics, 0)
+#endif
 EndPredDefs
