@@ -1,4 +1,4 @@
-/*  $Id: pl-proc.c,v 1.162 2007/05/16 09:42:27 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -148,6 +148,9 @@ resetProcedure(Procedure proc, bool isnew)
   { def->indexCardinality = 0;
     if ( def->functor->arity == 0 )
     { def->indexPattern = 0x0;
+    } else if ( true(def, DYNAMIC) )
+    { def->indexPattern = 0x1;
+      set(def, AUTOINDEX);
     } else
     { def->indexPattern = (0x0 | NEED_REINDEX);
       set(def, AUTOINDEX);
@@ -240,7 +243,8 @@ static int
 checkModifySystemProc(functor_t fd)
 { Procedure proc;
 
-  if ( (proc = isStaticSystemProcedure(fd)) )
+  if ( (proc = isStaticSystemProcedure(fd)) &&
+       true(proc->definition, P_ISO) )
     return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
 
   succeed;
@@ -265,29 +269,24 @@ lookupProcedureToDefine(functor_t def, Module m)
 get_functor() translates term  of  the   format  +Name/+Arity  into  the
 internal functor represenation. It fails and  raises an exception on the
 various possible format or represenation errors.  ISO compliant.
-
-The return value is 1 normally, -1  if no functor exists and GF_EXISTING
-is defined, and 0 if an error was raised.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define GF_EXISTING	1
-#define GF_PROCEDURE	2		/* check for max arity */
-
 static int
-get_arity(term_t t, int maxarity, int *arity)
+get_arity(term_t t, int extra, int maxarity, int *arity)
 { int a;
 
   if ( !PL_get_integer_ex(t, &a) )
     fail;
   if ( a < 0 )
     return PL_error(NULL, 0, NULL, ERR_DOMAIN,
-		    ATOM_not_less_than_zero, a);
+		    ATOM_not_less_than_zero, t);
+  a += extra;
   if ( maxarity >= 0 && a > maxarity )
   { char buf[100];
 
     return PL_error(NULL, 0,
 		    tostr(buf, "limit is %d, request = %d",
-			  maxarity, arity),
+			  maxarity, a),
 		    ERR_REPRESENTATION, ATOM_max_arity);
   }
 
@@ -297,23 +296,28 @@ get_arity(term_t t, int maxarity, int *arity)
 }
 
 
-static int
+int
 get_functor(term_t descr, functor_t *fdef, Module *m, term_t h, int how)
 { GET_LD
   term_t head = PL_new_term_ref();
+  int dcgpi=FALSE;
 
   PL_strip_module(descr, m, head);
 
-  if ( PL_is_functor(head, FUNCTOR_divide2) )
+  if ( PL_is_functor(head, FUNCTOR_divide2) ||
+       (dcgpi=PL_is_functor(head, FUNCTOR_div2)) )
   { term_t a = PL_new_term_ref();
     atom_t name;
-    int arity;
+    int arity = 0;
 
     _PL_get_arg(1, head, a);
     if ( !PL_get_atom_ex(a, &name) )
       fail;
     _PL_get_arg(2, head, a);
-    if ( !get_arity(a, (how&GF_PROCEDURE) ? MAXARITY : -1, &arity ) )
+    if ( !get_arity(a, 
+		    (dcgpi ? 2 : 0),
+		    (how&GF_PROCEDURE) ? MAXARITY : -1,
+		    &arity ) )
       fail;
     *fdef = PL_new_functor(name, arity);
     if ( h )
@@ -561,6 +565,7 @@ pl_current_predicate1(term_t spec, control_t ctx)
   term_t mt = 0;			/* module-term */
   term_t nt = 0;			/* name-term */
   term_t at = 0;			/* arity-term */
+  unsigned int aextra = 0;
 
   if ( ForeignControl(ctx) != FRG_CUTTED )
   { term_t pi = PL_copy_term_ref(spec);
@@ -578,7 +583,10 @@ pl_current_predicate1(term_t spec, control_t ctx)
     if ( PL_is_functor(pi, FUNCTOR_divide2) )
     { _PL_get_arg(1, pi, nt);
       _PL_get_arg(2, pi, at);
-
+    } else if ( PL_is_functor(pi, FUNCTOR_div2) )
+    { _PL_get_arg(1, pi, nt);
+      _PL_get_arg(2, pi, at);
+      aextra = 2;
     } else if ( PL_is_variable(pi) )
     { term_t a = PL_new_term_ref();
 
@@ -597,7 +605,13 @@ pl_current_predicate1(term_t spec, control_t ctx)
       { if ( !PL_is_variable(nt) )
 	  goto typeerror;
       }
-      if ( !PL_get_integer(at, &e->arity) )
+      if ( PL_get_integer(at, &e->arity) )
+      { if ( e->arity < 0 )
+	  return PL_error(NULL, 0, NULL, ERR_DOMAIN,
+			  ATOM_not_less_than_zero, at);
+
+	e->arity += aextra;
+      } else
       { if ( !PL_is_variable(at) )
 	  goto typeerror;
 	e->arity = -1;
@@ -609,17 +623,19 @@ pl_current_predicate1(term_t spec, control_t ctx)
       if ( mt )
       { atom_t mname;
 
-	if ( PL_get_atom(mt, &mname) )
-	{ e->module = isCurrentModule(mname);
-	  if ( !e->module )
-	    fail;
-	} else if ( PL_is_variable(mt) )
+	if ( PL_is_variable(mt) )
 	{ e->emod = newTableEnum(GD->tables.modules);
 
 	  if ( (sm = advanceTableEnum(e->emod)) )
 	    e->module = sm->value;
 	  else
 	    fail;			/* no modules!? */
+	} else if ( PL_get_atom_ex(mt, &mname) )
+	{ e->module = isCurrentModule(mname);
+	  if ( !e->module )
+	    fail;
+	} else
+	{ fail;
 	}
       } else
       { if ( environment_frame )
@@ -686,13 +702,14 @@ pl_current_predicate1(term_t spec, control_t ctx)
     
 	if ( (!e->name     || e->name == fd->name) &&
 	     (e->arity < 0 || (unsigned int)e->arity == fd->arity) &&
+	     fd->arity >= aextra &&
 	     isDefinedProcedure(proc) )
 	{ if ( mt )
 	    PL_unify_atom(mt, e->module->name);
 	  if ( !e->name )
 	    PL_unify_atom(nt, fd->name);
 	  if ( e->arity < 0 )
-	    PL_unify_integer(at, fd->arity);
+	    PL_unify_integer(at, fd->arity-aextra);
 
 	  ForeignRedoPtr(e);
 	}
@@ -819,7 +836,9 @@ assertProcedure(Procedure proc, Clause clause, int where ARG_LD)
   } else
   { if ( def->number_of_clauses == 25 &&
 	 true(def, AUTOINDEX) )
+    { DEBUG(2, Sdprintf("Request re-index for %s\n", predicateName(def)));
       def->indexPattern |= NEED_REINDEX;
+    }
   }
   UNLOCKDEF(def);
 
@@ -1035,7 +1054,7 @@ freeClause(Clause c ARG_LD)
 
   GD->statistics.codes -= c->code_size;
 #ifdef O_ATOMGC
-  unregisterAtomsClause(c);
+  forAtomsInClause(c, PL_unregister_atom);
 #endif
   freeHeap(c, sizeofClause(c->code_size));
 }
@@ -1253,9 +1272,9 @@ pl_garbage_collect_clauses(void)
     for(c=GD->procedures.dirty; c; c=c->next)
     { Definition def = c->definition;
 
-      assert(false(def, DYNAMIC));
       assert(true(def, P_DIRTYREG));
-      assert(def->references == 0);
+      if ( false(def, DYNAMIC) )
+	assert(def->references == 0);
     }
 
     markPredicatesInEnvironments(LD);
@@ -1272,20 +1291,24 @@ pl_garbage_collect_clauses(void)
       
       next = cell->next;
 
-      if ( def->references )
-      { assert(def->references == 1);
-	def->references = 0;
-	last = cell;
-      } else
-      { DEBUG(1, Sdprintf("gcClausesDefinition(%s)\n", predicateName(def)));
-	gcClausesDefinition(def);
-	clear(def, P_DIRTYREG);
-	freeHeap(cell, sizeof(*cell));
-	if ( last )
-	  last->next = next;
-	else
-	  GD->procedures.dirty = next;
+      if ( false(def, DYNAMIC) )
+      { if ( def->references )
+	{ assert(def->references == 1);
+	  def->references = 0;
+	  last = cell;
+	  continue;
+	} else
+	{ DEBUG(1, Sdprintf("gcClausesDefinition(%s)\n", predicateName(def)));
+	  gcClausesDefinition(def);
+	}
       }
+  
+      clear(def, P_DIRTYREG);
+      freeHeap(cell, sizeof(*cell));
+      if ( last )
+	last->next = next;
+      else
+	GD->procedures.dirty = next;
     }
 
 #ifdef O_PLMT
@@ -1399,7 +1422,7 @@ Definition
 autoImport(functor_t f, Module m)
 { GET_LD
   Procedure proc;
-  Definition def;
+  Definition def, odef;
   ListCell c;
 					/* Defined: no problem */
   if ( (proc = isCurrentProcedure(f, m)) && isDefinedProcedure(proc) )
@@ -1416,18 +1439,20 @@ found:
     proc = lookupProcedure(f, m);
 					/* Safe? See above */
 					/* TBD: find something better! */
+  odef = proc->definition;
+  proc->definition = def;
+
 #ifdef O_PLMT
   PL_LOCK(L_THREAD);
   if ( (GD->statistics.threads_created -
 	GD->statistics.threads_finished) == 1 )
   { assert(false(proc->definition, P_DIRTYREG));
-    freeHeap(proc->definition, sizeof(struct definition));
+    freeHeap(odef, sizeof(struct definition));
   }
   PL_UNLOCK(L_THREAD);
 #else
-  freeHeap(proc->definition, sizeof(struct definition));
+  freeHeap(odef, sizeof(struct definition));
 #endif
-  proc->definition = def;
 
   return def;
 }
@@ -1545,14 +1570,10 @@ trapUndefined_unlocked(LocalFrame *frp, Code PC, Procedure proc ARG_LD)
 				/* No one wants to intercept */
 error:
   if ( GD->bootsession )
-    sysError("Undefined predicate: %s", predicateName(def));
-  else if ( true(module, UNKNOWN_ERROR) )
-  { Definition caller;
-    if ( (*frp)->parent )
-      caller = (*frp)->parent->predicate;
-    else
-      caller = NULL;
-    PL_error(NULL, 0, NULL, ERR_UNDEFINED_PROC, def, caller);
+  { sysError("Undefined predicate: %s", predicateName(def));
+  } else if ( true(module, UNKNOWN_ERROR) )
+  {			/* caller is in environment frame, which is default */
+    PL_error(NULL, 0, NULL, ERR_UNDEFINED_PROC, def, NULL);
   } else
   { fid_t fid = PL_open_foreign_frame();
     term_t pred = PL_new_term_ref();
@@ -1804,10 +1825,10 @@ do_abolish(Module m, term_t atom, term_t arity)
   functor_t f;
   Procedure proc;
   atom_t name;
-  int a;
+  int a = 0;
 
   if ( !PL_get_atom_ex(atom, &name) ||
-       !get_arity(arity, MAXARITY, &a) )
+       !get_arity(arity, 0, MAXARITY, &a) )
     fail;
 
   if ( !(f = isCurrentFunctor(name, a)) )
@@ -1876,6 +1897,7 @@ attribute_mask(atom_t key)
   if (key == ATOM_volatile)	 return VOLATILE;
   if (key == ATOM_thread_local)  return P_THREAD_LOCAL;
   if (key == ATOM_noprofile)     return P_NOPROFILE;
+  if (key == ATOM_iso)      	 return P_ISO;
 
   return 0;
 }
@@ -2034,6 +2056,9 @@ setDynamicProcedure(Procedure proc, bool isdyn)
 
       return PL_error(NULL, 0, msg,
 		      ERR_MODIFY_STATIC_PROC, proc);
+    } else if ( def->functor->arity > 0 )
+    { def->indexPattern = 0x1;
+      set(def, AUTOINDEX);
     }
   ok:
     set(def, DYNAMIC);
@@ -2144,7 +2169,7 @@ pl_set_predicate_attribute(term_t pred,
   { DEBUG(2, Sdprintf("Associating %s to %s (%p)\n",
 		      predicateName(def), PL_atom_chars(source_file_name),
 		      def));
-    addProcedureSourceFile(lookupSourceFile(source_file_name), proc);
+    addProcedureSourceFile(lookupSourceFile(source_file_name, TRUE), proc);
 
     if ( SYSTEM_MODE )
     { set(def, SYSTEM|HIDE_CHILDS);
@@ -2376,7 +2401,7 @@ cleanupSourceFiles(void)
 
 
 SourceFile
-lookupSourceFile(atom_t name)
+lookupSourceFile(atom_t name, int create)
 { SourceFile file;
   Symbol s;
 
@@ -2386,7 +2411,7 @@ lookupSourceFile(atom_t name)
 
   if ( (s=lookupHTable(sourceTable, (void*)name)) )
   { file = s->value;
-  } else
+  } else if ( create )
   { GET_LD
 
     file = (SourceFile) allocHeap(sizeof(struct sourceFile));
@@ -2402,6 +2427,8 @@ lookupSourceFile(atom_t name)
     registerSourceFile(file);
 
     addHTable(sourceTable, (void*)name, file);
+  } else
+  { file = NULL;
   }
   UNLOCK();
 
@@ -2461,7 +2488,7 @@ addProcedureSourceFile(SourceFile sf, Procedure proc)
 
 
 void
-redefineProcedure(Procedure proc, SourceFile sf)
+redefineProcedure(Procedure proc, SourceFile sf, unsigned int suppress)
 { GET_LD
   Definition def = proc->definition;
 
@@ -2478,7 +2505,7 @@ redefineProcedure(Procedure proc, SourceFile sf)
   { ClauseRef first = hasClausesDefinition(def);
 
     if ( first && first->clause->source_no == sf->index )
-    { if ( (debugstatus.styleCheck & DISCONTIGUOUS_STYLE) &&
+    { if ( ((debugstatus.styleCheck & ~suppress) & DISCONTIGUOUS_STYLE) &&
 	   false(def, DISCONTIGUOUS) )
 	printMessage(ATOM_warning,
 		     PL_FUNCTOR_CHARS, "discontiguous", 1,
@@ -2546,7 +2573,7 @@ pl_source_file(term_t descr, term_t file, control_t h)
     succeed;
 
   if ( !PL_get_atom_ex(file, &name) ||
-       !(sf = lookupSourceFile(name)) )
+       !(sf = lookupSourceFile(name, FALSE)) )
     fail;
 
   switch( ForeignControl(h) )
@@ -2581,34 +2608,43 @@ pl_source_file(term_t descr, term_t file, control_t h)
   fail;
 }
 
-
-word
-pl_time_source_file(term_t file, term_t time, control_t h)
-{ GET_LD
+static
+PRED_IMPL("$time_source_file", 3, time_source_file, PL_FA_NONDETERMINISTIC)
+{ PRED_LD
   int index;
   int mx = (int)entriesBuffer(&GD->files.source_files, SourceFile);
+  term_t file = A1;
+  term_t time = A2;
+  term_t type = A3;			/* user or system */
+  fid_t fid;
 
-  switch( ForeignControl(h) )
+  switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
       index = 0;
       break;
     case FRG_REDO:
-      index = (int)ForeignContextInt(h);
+      index = (int)CTX_INT;
       break;
     case FRG_CUTTED:
     default:
       succeed;
   }
 
+  fid = PL_open_foreign_frame();
   for(; index < mx; index++)
   { SourceFile f = fetchBuffer(&GD->files.source_files, index, SourceFile);
 
-    if ( !f->system )
-    { if ( PL_unify_atom(file, f->name) && unifyTime(time, f->time) )
-	ForeignRedoInt(index+1);
+    if ( PL_unify_atom(file, f->name) &&
+	 unifyTime(time, f->time) &&
+	 PL_unify_atom(type, f->system ? ATOM_system : ATOM_user) )
+    { PL_close_foreign_frame(fid);
+      ForeignRedoInt(index+1);
     }
+
+    PL_rewind_foreign_frame(fid);
   }
 
+  PL_close_foreign_frame(fid);
   fail;
 }
 
@@ -2715,7 +2751,7 @@ pl_start_consult(term_t file)
   atom_t name;
 
   if ( PL_get_atom(file, &name) )
-  { SourceFile f = lookupSourceFile(name);
+  { SourceFile f = lookupSourceFile(name, TRUE);
 
     f->time = LastModifiedFile(stringAtom(name));
     startConsult(f);
@@ -2729,20 +2765,23 @@ pl_start_consult(term_t file)
 		 *       DEBUGGER SUPPORT	*
 		 *******************************/
 
-word
-pl_clause_from_source(term_t file, term_t line, term_t clause)
-{ GET_LD
+static
+PRED_IMPL("$clause_from_source", 3, clause_from_source, 0)
+{ PRED_LD
   atom_t name;
   SourceFile f;
   int ln;
   ListCell cell;
   Clause c = NULL;
 
-  if ( !PL_get_atom(file, &name) ||
-       !(f = lookupSourceFile(name)) ||
-       !PL_get_integer(line, &ln) )
-    return warning("clause_from_source/3: instantiation fault");
-  
+  term_t file = A1;
+  term_t line = A2;
+  term_t clause = A3;
+
+  if ( !PL_get_atom_ex(file, &name) ||
+       !(f = lookupSourceFile(name, FALSE)) ||
+       !PL_get_integer_ex(line, &ln) )
+    fail;
 
   for(cell = f->procedures; cell; cell = cell->next)
   { Procedure proc = cell->value;
@@ -2908,3 +2947,13 @@ pl_list_generations(term_t desc)
 
 
 #endif /*O_MAINTENANCE*/
+
+
+		 /*******************************
+		 *      PUBLISH PREDICATES	*
+		 *******************************/
+
+BeginPredDefs(proc)
+  PRED_DEF("$time_source_file", 3, time_source_file, PL_FA_NONDETERMINISTIC)
+  PRED_DEF("$clause_from_source", 3, clause_from_source, 0)
+EndPredDefs
