@@ -1,4 +1,4 @@
-/*  $Id: pl-stream.c,v 1.163 2007/07/02 20:33:54 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -140,11 +140,18 @@ STRYLOCK(IOSTREAM *s)
 #endif
 
 #include "pl-error.h"
+typedef void *record_t;
+typedef intptr_t term_t;
+
 extern int 			fatalError(const char *fm, ...);
 extern int 			PL_error(const char *pred, int arity,
 					 const char *msg, int id, ...);
 extern int			PL_handle_signals();
 extern IOENC			initEncoding(void);
+extern int			reportStreamError(IOSTREAM *s);
+extern record_t			PL_record(term_t t);
+extern int			PL_thread_self(void);
+
 
 		 /*******************************
 		 *	      BUFFER		*
@@ -158,28 +165,71 @@ case, but this is only used by the output stream Svfprintf() where it is
 not needed.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
-S__setbuf(IOSTREAM *s, char *buffer, int size)
-{ if ( size == 0 )
+static size_t
+S__setbuf(IOSTREAM *s, char *buffer, size_t size)
+{ char *newbuf, *newunbuf;
+  int newflags = s->flags;
+
+  if ( size == 0 )
     size = SIO_BUFSIZE;
 
-  S__removebuf(s);
-  s->bufsize = size;
+  if ( (s->flags & SIO_OUTPUT) )
+  { if ( S__removebuf(s) < 0 )
+      return -1;
+  }
 
   if ( buffer )
-  { s->unbuffer = s->buffer = buffer;
-    s->flags |= SIO_USERBUF;
+  { newunbuf = newbuf = buffer;
+    newflags |= SIO_USERBUF;
   } else
-  { if ( !(s->unbuffer = malloc(s->bufsize+UNDO_SIZE)) )
+  { if ( !(newunbuf = malloc(size+UNDO_SIZE)) )
     { errno = ENOMEM;
       return -1;
     }
-    s->flags &= ~SIO_USERBUF;
-    s->buffer = s->unbuffer + UNDO_SIZE;
+    newflags &= ~SIO_USERBUF;
+    newbuf = newunbuf + UNDO_SIZE;
   }
 
-  s->limitp = &s->buffer[s->bufsize];
-  s->bufp   = s->buffer;
+  if ( (s->flags & SIO_INPUT) )
+  { size_t buffered = s->limitp - s->bufp;
+    size_t copy = (buffered < size ? buffered : size);
+
+    if ( size < buffered )
+    { size_t offset = size - buffered;
+      int64_t newpos;
+
+      if ( s->functions->seek64 )
+      { newpos = (*s->functions->seek64)(s->handle, offset, SIO_SEEK_CUR);
+      } else if ( s->functions->seek )
+      { newpos = (*s->functions->seek)(s->handle, (long)offset, SIO_SEEK_CUR);
+      } else
+      { newpos = -1;
+	errno = ESPIPE;
+      }
+
+      if ( newpos == -1 )
+      { if ( !(newflags & SIO_USERBUF) )
+	{ int oldeno = errno;
+
+	  free(newunbuf);
+	  errno = oldeno;
+	  return -1;
+	}
+      }
+    }
+
+    memcpy(newbuf, s->bufp, copy);
+    S__removebuf(s);
+    s->unbuffer = newunbuf;
+    s->bufp = s->buffer = newbuf;
+    s->limitp = s->buffer+copy;
+  } else
+  { s->unbuffer = newunbuf;
+    s->bufp = s->buffer = newbuf;
+    s->limitp = &s->buffer[size];
+  }
+  s->bufsize = (int)size;
+  s->flags = newflags;
 
   return size;
 }
@@ -227,7 +277,7 @@ Sname(IOSTREAM *s)
 
 static void
 print_trace(void)
-{ void *array[3];
+{ void *array[7];
   size_t size;
   char **strings;
   size_t i;
@@ -237,7 +287,7 @@ print_trace(void)
      
   printf(" Stack:");
   for(i = 1; i < size; i++)
-  { printf(" [%d] %s", i, strings[i]);
+  { printf("\n\t[%ld] %s", (long)i, strings[i]);
   }
   printf("\n");
        
@@ -251,13 +301,15 @@ Slock(IOSTREAM *s)
 { SLOCK(s);
 
 #ifdef DEBUG_IO_LOCKS
-  printf("  Lock: %d: %s: %d locks", PL_thread_self(), Sname(s), s->locks+1);
-  print_trace();
+  if ( s->locks > 2 )
+  { printf("  Lock [%d]: %s: %d locks", PL_thread_self(), Sname(s), s->locks+1);
+    print_trace();
+  }
 #endif
 
   if ( !s->locks++ )
   { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
-      return S__setbuf(s, NULL, TMPBUFSIZE);
+      return S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 ? -1 : 0;
   }
 
   return 0;
@@ -271,20 +323,22 @@ StryLock(IOSTREAM *s)
 
   if ( !s->locks++ )
   { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
-      return S__setbuf(s, NULL, TMPBUFSIZE);
+      return S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 ? -1 : 0;
   }
 
   return 0;
 }
 
 
-int
-Sunlock(IOSTREAM *s)
+static int
+S__unlock(IOSTREAM *s)
 { int rval = 0;
 
 #ifdef DEBUG_IO_LOCKS
-  printf("Unlock: %d: %s: %d locks", PL_thread_self(), Sname(s), s->locks-1);
-  print_trace();
+  if ( s->locks > 3 )
+  { printf("Unlock [%d]: %s: %d locks", PL_thread_self(), Sname(s), s->locks-1);
+    print_trace();
+  }
 #endif
 
   if ( s->locks )
@@ -296,6 +350,13 @@ Sunlock(IOSTREAM *s)
   { assert(0);
   }
 
+  return rval;
+}
+
+
+int
+Sunlock(IOSTREAM *s)
+{ int rval = S__unlock(s);
   SUNLOCK(s);
 
   return rval;
@@ -365,7 +426,7 @@ S__flushbufc(int c, IOSTREAM *s)
 	c = -1;
       }
     } else
-    { if ( S__setbuf(s, NULL, 0) < 0 )
+    { if ( S__setbuf(s, NULL, 0) == (size_t)-1 )
       { s->flags |= SIO_FERR;
 	c = -1;
       } else
@@ -459,7 +520,7 @@ S__fillbuf(IOSTREAM *s)
     size_t len;
 
     if ( !s->buffer )
-    { if ( S__setbuf(s, NULL, 0) < 0 )
+    { if ( S__setbuf(s, NULL, 0) == (size_t)-1 )
 	return -1;
       s->bufp = s->limitp = s->buffer;
       len = s->bufsize;
@@ -516,10 +577,6 @@ update_linepos(IOSTREAM *s, int c)
       p->lineno++;
       p->linepos = 0;
       s->flags &= ~SIO_NOLINEPOS;
-#ifdef __WINDOWS__
-      if ( s->flags & O_TEXT )
-	p->charno++;			/* writes one extra! */
-#endif
       break;
     case '\r':
       p->linepos = 0;
@@ -540,6 +597,14 @@ update_linepos(IOSTREAM *s, int c)
 
 
 
+int
+S__fcheckpasteeof(IOSTREAM *s, int c)
+{ S__checkpasteeof(s, c);
+
+  return c;
+}
+
+
 int
 S__fupdatefilepos_getc(IOSTREAM *s, int c)
 { IOPOS *p = s->position;
@@ -560,6 +625,7 @@ S__updatefilepos(IOSTREAM *s, int c)
   { update_linepos(s, c);
     p->charno++;
   }
+  S__checkpasteeof(s,c);
 
   return c;
 }
@@ -626,9 +692,11 @@ unget_byte(int c, IOSTREAM *s)
 
   *--s->bufp = c;
   if ( p )
-  { p->charno--;
+  { p->charno--;			/* FIXME: not correct */
     p->byteno--;
-    s->flags |= (SIO_NOLINENO|SIO_NOLINEPOS);
+    if ( c == '\n' )
+      p->lineno--;
+    s->flags |= SIO_NOLINEPOS;
   }
 }
 
@@ -673,22 +741,9 @@ reperror(int c, IOSTREAM *s)
 
 
 
-int
-Sputcode(int c, IOSTREAM *s)
-{ if ( c < 0 )
-    return reperror(c, s);
-
-  if ( s->tee && s->tee->magic == SIO_MAGIC )
-    Sputcode(c, s->tee);
-
-#ifdef CRLF_MAPPING
-  if ( c == '\n' && (s->flags&SIO_TEXT) )
-  { if ( Sputcode('\r', s) < 0 )
-      return -1;
-  }
-#endif
-
-  switch(s->encoding)
+static int
+put_code(int c, IOSTREAM *s)
+{ switch(s->encoding)
   { case ENC_OCTET:
     case ENC_ISO_LATIN_1:
       if ( c >= 256 )
@@ -786,6 +841,23 @@ Sputcode(int c, IOSTREAM *s)
 }
 
 
+int
+Sputcode(int c, IOSTREAM *s)
+{ if ( c < 0 )
+    return reperror(c, s);
+
+  if ( s->tee && s->tee->magic == SIO_MAGIC )
+    Sputcode(c, s->tee);
+
+  if ( c == '\n' && (s->flags&SIO_TEXT) && s->newline == SIO_NL_DOS )
+  { if ( put_code('\r', s) < 0 )
+      return -1;
+  }
+
+  return put_code(c, s);
+}
+
+
 int
 Scanrepresent(int c, IOSTREAM *s)
 { switch(s->encoding)
@@ -829,10 +901,7 @@ int
 Sgetcode(IOSTREAM *s)
 { int c;
 
-#ifdef CRLF_MAPPING
 retry:
-#endif
-
   switch(s->encoding)
   { case ENC_OCTET:
     case ENC_ISO_LATIN_1:
@@ -858,8 +927,8 @@ retry:
       for(;;)
       { if ( (c = get_byte(s)) == EOF )
 	{ if ( n == 0 )
-	    return EOF;
-	  else
+	  { goto out;
+	  } else
 	  { Sseterr(s, SIO_WARN, "EOF in multibyte Sequence");
 	    goto mberr;
 	  }
@@ -916,7 +985,9 @@ retry:
 
       c1 = get_byte(s);
       if ( c1 == EOF )
-	return EOF;
+      { c = -1;
+	goto out;
+      }
       c2 = get_byte(s);
 
       if ( c2 == EOF )
@@ -941,7 +1012,8 @@ retry:
 
 	if ( c1 == EOF )
 	{ if ( n == 0 )
-	  { return EOF;
+	  { c = -1;
+	    goto out;
 	  } else
 	  { Sseterr(s, SIO_WARN, "EOF in UCS character");
 	    c = UTF8_MALFORMED_REPLACEMENT; 
@@ -961,10 +1033,15 @@ retry:
   }
 
 out:
-#ifdef CRLF_MAPPING
   if ( c == '\r' && (s->flags&SIO_TEXT) )
-    goto retry;
-#endif
+  { switch(s->newline)
+    { case SIO_NL_DETECT:
+	s->newline = SIO_NL_DOS;
+        /*FALLTHROUGH*/
+      case SIO_NL_DOS:
+	goto retry;
+    }
+  }
 
   if ( s->tee && s->tee->magic == SIO_MAGIC && c != -1 )
     Sputcode(c, s->tee);
@@ -1136,13 +1213,40 @@ Sfread(void *data, size_t size, size_t elms, IOSTREAM *s)
 { size_t chars = size * elms;
   char *buf = data;
 
-  for( ; chars > 0; chars-- )
-  { int c;
+  if ( s->position )
+  { for( ; chars > 0; chars-- )
+    { int c;
 
-    if ( (c = Sgetc(s)) == EOF )
-      break;
+      if ( (c = Sgetc(s)) == EOF )
+	break;
+      
+      *buf++ = c & 0xff;
+    }
+  } else 
+  { while(chars > 0)
+    { int c;
 
-    *buf++ = c & 0xff;
+      if ( s->bufp < s->limitp )
+      { size_t avail = s->limitp - s->bufp;
+	
+	if ( chars <= avail )
+	{ memcpy(buf, s->bufp, chars);
+	  s->bufp += chars;
+	  return elms;
+	} else
+	{ memcpy(buf, s->bufp, avail);
+	  chars -= avail;
+	  buf += avail;
+	  s->bufp += avail;
+	}
+      }
+      
+      if ( (c = S__fillbuf(s)) == EOF )
+	break;
+
+      *buf++ = c & 0xff;
+      chars--;
+    }
   }
   
   return (size*elms - chars)/size;
@@ -1176,7 +1280,10 @@ Sread_pending(IOSTREAM *s, char *buf, size_t limit, int flags)
   { int c = S__fillbuf(s);
 
     if ( c < 0 )
+    { if ( (s->flags & SIO_FEOF) )
+	return 0;
       return c;
+    }
 
     buf[0] = c;
     limit--;
@@ -1345,6 +1452,13 @@ Sseterr(IOSTREAM *s, int flag, const char *message)
 }
 
 
+void
+Sset_exception(IOSTREAM *s, term_t ex)
+{ s->exception = PL_record(ex);
+  s->flags |= SIO_FERR;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Set the encoding of a stream. The enc   argument is the new encoding. If
 old is not NULL, the old encoding is written to the given location.
@@ -1514,7 +1628,7 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
   }
 
 update:
-  s->flags &= ~SIO_FEOF;		/* not on eof of file anymore */
+  s->flags &= ~(SIO_FEOF|SIO_FEOF2);	/* not on eof of file anymore */
 
   if ( s->position )
   { s->flags |= (SIO_NOLINENO|SIO_NOLINEPOS); /* no update this */
@@ -1602,7 +1716,8 @@ Sclose(IOSTREAM *s)
     return rval;
 
   if ( s->upstream )
-  { errno = EPERM;
+  { Sseterr(s, SIO_FERR, "Locked by upstream filter");
+    reportStreamError(s);
     return -1;
   }
 
@@ -1623,10 +1738,18 @@ Sclose(IOSTREAM *s)
   }
 #endif
   if ( s->functions->close && (*s->functions->close)(s->handle) < 0 )
+  { s->flags |= SIO_FERR;
     rval = -1;
-  run_close_hooks(s);
+  }
   while(s->locks > 0)			/* remove buffer-locks */
-    rval = Sunlock(s);
+  { int rc = S__unlock(s);
+
+    if ( rval == 0 )
+      rval = rc;
+  }
+  if ( rval < 0 )
+    reportStreamError(s);
+  run_close_hooks(s);			/* deletes Prolog registration */
 
   SUNLOCK(s);
 
@@ -2514,6 +2637,20 @@ IOFUNCTIONS Sfilefunctions =
 };
 
 
+IOFUNCTIONS Sttyfunctions =
+{ Sread_file,
+  Swrite_file,
+  NULL,
+  Sclose_file,
+  Scontrol_file,
+#ifdef O_LARGEFILES
+  Sseek_file64
+#else
+  NULL
+#endif
+};
+
+
 IOSTREAM *
 Snew(void *handle, int flags, IOFUNCTIONS *functions)
 { IOSTREAM *s;
@@ -2532,6 +2669,9 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   s->timeout       = -1;		/* infinite */
   s->posbuf.lineno = 1;
   s->encoding      = ENC_ISO_LATIN_1;
+#if CRLF_MAPPING
+  s->newline       = SIO_NL_DOS;
+#endif
   if ( flags & SIO_RECORDPOS )
     s->position = &s->posbuf;
 #ifdef O_PLMT
@@ -2569,7 +2709,7 @@ IOSTREAM *
 Sopen_file(const char *path, const char *how)
 { int fd;
   int oflags = O_BINARY;
-  int flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
+  int flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS|SIO_FBUF;
   int op = *how++;
   intptr_t lfd;
   enum {lnone=0,lread,lwrite} lock = lnone;
@@ -2694,9 +2834,9 @@ Sfdopen(int fd, const char *type)
 #endif
 
   if ( *type == 'r' )
-    flags = SIO_FILE|SIO_INPUT|SIO_RECORDPOS;
+    flags = SIO_FILE|SIO_INPUT|SIO_RECORDPOS|SIO_FBUF;
   else
-    flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS;
+    flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS|SIO_FBUF;
 
   lfd = (intptr_t)fd;
 
@@ -2799,9 +2939,9 @@ Sopen_pipe(const char *command, const char *type)
   { int flags;
 
     if ( *type == 'r' )
-      flags = SIO_PIPE|SIO_INPUT;
+      flags = SIO_PIPE|SIO_INPUT|SIO_FBUF;
     else
-      flags = SIO_PIPE|SIO_OUTPUT;
+      flags = SIO_PIPE|SIO_OUTPUT|SIO_FBUF;
 
     return Snew((void *)fd, flags, &Spipefunctions);
   }
@@ -3126,7 +3266,7 @@ Sopen_string(IOSTREAM *s, char *buf, size_t size, const char *mode)
 
 #define STDIO(n, f) { NULL, NULL, NULL, NULL, \
 		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
-		      ((void *)(n)), &Sfilefunctions, \
+		      ((void *)(n)), &Sttyfunctions, \
 		      0, NULL, \
 		      (void (*)(void *))0, NULL, \
 		      -1, \
@@ -3134,7 +3274,7 @@ Sopen_string(IOSTREAM *s, char *buf, size_t size, const char *mode)
 		      ENC_ISO_LATIN_1 \
 		    }
 
-#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
+#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY|SIO_TEXT)
 #define STDIO_STREAMS \
   STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
   STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPL), 	/* Soutput */ \
@@ -3161,17 +3301,24 @@ SinitStreams()
 
     for(i=0; i<=2; i++)
     { if ( !isatty(i) )
-	S__iob[i].flags &= ~SIO_ISATTY;
+      { S__iob[i].flags &= ~SIO_ISATTY;
+	S__iob[i].functions = &Sfilefunctions; /* Check for pipe? */
+      }
       if ( S__iob[i].encoding == ENC_ISO_LATIN_1 )
 	S__iob[i].encoding = enc;
 #ifdef O_PLMT
       S__iob[i].mutex = malloc(sizeof(recursiveMutex));
       recursiveMutexInit(S__iob[i].mutex);
 #endif
-#ifdef __WINDOWS__
-      pt_init();
+#if CRLF_MAPPING
+      _setmode(i, O_BINARY);
+      S__iob[i].newline = SIO_NL_DOS;
 #endif
     }
+
+#ifdef __WINDOWS__
+    pt_init();				/* init popen() issues */
+#endif
   }
 }
 
@@ -3229,11 +3376,17 @@ Sclosehook(void (*hook)(IOSTREAM *s))
 
 void
 Sreset(void)
-{ IOSTREAM *s = Sinput;
+{ IOSTREAM *s;
 
-  if ( s && s->magic == SIO_MAGIC )
+  if ( (s=Sinput) && s->magic == SIO_MAGIC )
   { s->bufp = s->limitp = s->buffer;
   }
+  if ( (s=Soutput) && s->magic == SIO_MAGIC )
+  { s->bufp = s->buffer;
+  }
+  if ( (s=Serror) && s->magic == SIO_MAGIC )
+  { s->bufp = s->buffer;
+  }
 } 
 
 
