@@ -2418,6 +2418,9 @@ thread(status-1) :-
 	Id2 == Id,
 	Status == true, !,
 	thread_join(Id2, _).
+thread(create_error-1) :-
+	catch(thread_create(true, _, [local(a)]), E, true),
+	E = error(domain_error(thread_option, local(a)), _).
 	
 
 		 /*******************************
